# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Application and entities that are used to describe components on Discord."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "ActionRowComponent",
    "ButtonComponent",
    "ButtonStyle",
    "ChannelSelectMenuComponent",
    "ComponentType",
    "ContainerComponent",
    "ContainerTypesT",
    "FileComponent",
    "InteractiveButtonTypes",
    "InteractiveButtonTypesT",
    "MediaGalleryComponent",
    "MediaGalleryItem",
    "MediaLoadingType",
    "MediaResource",
    "MessageActionRowComponent",
    "MessageComponentTypesT",
    "ModalActionRowComponent",
    "ModalComponentTypesT",
    "PartialComponent",
    "SectionComponent",
    "SelectMenuComponent",
    "SelectMenuOption",
    "SeparatorComponent",
    "SpacingType",
    "TextDisplayComponent",
    "TextInputComponent",
    "TextInputStyle",
    "TextSelectMenuComponent",
    "ThumbnailComponent",
    "TopLevelComponentTypesT",
)

import typing

import attrs

from hikari import channels
from hikari import emojis
from hikari import files
from hikari.internal import enums
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    import concurrent.futures

    from hikari import channels
    from hikari import colors
    from hikari import emojis
    from hikari import undefined


@typing.final
class ComponentType(int, enums.Enum):
    """Types of components found within Discord."""

    ACTION_ROW = 1
    """A non-interactive container component for other types of components.

    !!! note
        As of writing this can only contain one component type.
    """

    BUTTON = 2
    """A button component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    TEXT_SELECT_MENU = 3
    """A text select component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    TEXT_INPUT = 4
    """A text input component.

    !!! note
        This component may only be used inside a modal container.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    USER_SELECT_MENU = 5
    """A user select component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    ROLE_SELECT_MENU = 6
    """A role select component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    MENTIONABLE_SELECT_MENU = 7
    """A mentionable (users and roles) select component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    CHANNEL_SELECT_MENU = 8
    """A channel select component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.ACTION_ROW`][].
    """

    SECTION = 9
    """A section component.

    !!! note
        As this is a container component it can never be contained within another
        component and therefore will always be top-level.
    """

    TEXT_DISPLAY = 10
    """A text display component."""

    THUMBNAIL = 11
    """A thumbnail component.

    !!! note
        This cannot be top-level and must be within a container component such
        as [`hikari.components.ComponentType.SECTION`][].
    """

    MEDIA_GALLERY = 12
    """A media gallery component."""

    FILE = 13
    """A file component."""

    SEPARATOR = 14
    """A separator component."""

    CONTAINER = 17
    """A container component.

    !!! note
        As this is a container component it can never be contained within another
        component and therefore will always be top-level.
    """


@typing.final
class ButtonStyle(int, enums.Enum):
    """Enum of the available button styles.

    More information, such as how these look, can be found at
    https://discord.com/developers/docs/interactions/message-components#button-object-button-styles
    """

    PRIMARY = 1
    """A blurple "call to action" button."""

    SECONDARY = 2
    """A grey neutral button."""

    SUCCESS = 3
    """A green button."""

    DANGER = 4
    """A red button (usually indicates a destructive action)."""

    LINK = 5
    """A grey button which navigates to a URL.

    !!! warning
        Unlike the other button styles, clicking this one will not trigger an
        interaction and custom_id shouldn't be included for this style.
    """


@typing.final
class TextInputStyle(int, enums.Enum):
    """A text input style."""

    SHORT = 1
    """Intended for short single-line text."""

    PARAGRAPH = 2
    """Intended for much longer inputs."""


@typing.final
class SpacingType(int, enums.Enum):
    """Spacing Type.

    The type of spacing for a [SeparatorComponent][]
    """

    SMALL = 1
    """A small separator."""

    LARGE = 2
    """A large separator."""


@typing.final
class MediaLoadingType(int, enums.Enum):
    """Media loading type."""

    UNKNOWN = 0
    """Media is in an unknown loading state."""

    LOADING = 1
    """Media is loading."""

    LOADED_SUCCESS = 2
    """Media has successfully loaded."""

    LOADED_NOT_FOUND = 3
    """Media was not found."""


@attrs.define(kw_only=True, weakref_slot=False)
class PartialComponent:
    """Base class for all component entities."""

    type: ComponentType | int = attrs.field()
    """The type of component this is."""

    id: int = attrs.field()
    """The ID of the interaction."""


AllowedComponentsT = typing.TypeVar("AllowedComponentsT", bound="PartialComponent")


@attrs.define(weakref_slot=False)
class ActionRowComponent(typing.Generic[AllowedComponentsT], PartialComponent):
    """Represents a row of components."""

    components: typing.Sequence[AllowedComponentsT] = attrs.field()
    """Sequence of the components contained within this row."""

    @typing.overload
    def __getitem__(self, index: int, /) -> PartialComponent: ...

    @typing.overload
    def __getitem__(self, slice_: slice, /) -> typing.Sequence[AllowedComponentsT]: ...

    def __getitem__(self, index_or_slice: int | slice, /) -> PartialComponent | typing.Sequence[AllowedComponentsT]:
        return self.components[index_or_slice]

    def __iter__(self) -> typing.Iterator[AllowedComponentsT]:
        return iter(self.components)

    def __len__(self) -> int:
        return len(self.components)


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ButtonComponent(PartialComponent):
    """Represents a button component."""

    style: ButtonStyle | int = attrs.field(eq=False)
    """The button's style."""

    label: str | None = attrs.field(eq=False)
    """Text label which appears on the button."""

    emoji: emojis.Emoji | None = attrs.field(eq=False)
    """Custom or unicode emoji which appears on the button."""

    custom_id: str | None = attrs.field(hash=True)
    """Developer defined identifier for this button (will be <= 100 characters).

    !!! note
        This is required for the following button styles:

        * [`hikari.components.ButtonStyle.PRIMARY`][]
        * [`hikari.components.ButtonStyle.SECONDARY`][]
        * [`hikari.components.ButtonStyle.SUCCESS`][]
        * [`hikari.components.ButtonStyle.DANGER`][]
    """

    url: str | None = attrs.field(eq=False)
    """Url for [`hikari.components.ButtonStyle.LINK`][] style buttons."""

    is_disabled: bool = attrs.field(eq=False)
    """Whether the button is disabled."""


@attrs.define(kw_only=True, weakref_slot=False)
class SelectMenuOption:
    """Represents an option for a [`hikari.components.SelectMenuComponent`][]."""

    label: str = attrs.field()
    """User-facing name of the option, max 100 characters."""

    value: str = attrs.field()
    """Dev-defined value of the option, max 100 characters."""

    description: str | None = attrs.field()
    """Optional description of the option, max 100 characters."""

    emoji: emojis.Emoji | None = attrs.field(eq=False)
    """Custom or unicode emoji which appears on the button."""

    is_default: bool = attrs.field()
    """Whether this option will be selected by default."""


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class SelectMenuComponent(PartialComponent):
    """Represents a select menu component."""

    custom_id: str = attrs.field(hash=True)
    """Developer defined identifier for this menu (will be <= 100 characters)."""

    placeholder: str | None = attrs.field(eq=False)
    """Custom placeholder text shown if nothing is selected, max 100 characters."""

    min_values: int = attrs.field(eq=False)
    """The minimum amount of options which must be chosen for this menu.

    This will be greater than or equal to 0 and will be less than or equal to
    [`hikari.components.SelectMenuComponent.max_values`][].
    """

    max_values: int = attrs.field(eq=False)
    """The minimum amount of options which can be chosen for this menu.

    This will be less than or equal to 25 and will be greater than or equal to
    [`hikari.components.SelectMenuComponent.min_values`][].
    """

    is_disabled: bool = attrs.field(eq=False)
    """Whether the select menu is disabled."""


@attrs.define(kw_only=True, weakref_slot=False)
class TextSelectMenuComponent(SelectMenuComponent):
    """Represents a text select menu component."""

    options: typing.Sequence[SelectMenuOption] = attrs.field(eq=False)
    """Sequence of up to 25 of the options set for this menu."""


@attrs.define(kw_only=True, weakref_slot=False)
class ChannelSelectMenuComponent(SelectMenuComponent):
    """Represents a channel select menu component."""

    channel_types: typing.Sequence[int | channels.ChannelType] = attrs.field(eq=False)
    """The valid channel types for this menu."""


@attrs.define(kw_only=True, weakref_slot=False)
class TextInputComponent(PartialComponent):
    """Represents a text input component."""

    custom_id: str = attrs.field(repr=True)
    """Developer set custom ID used for identifying interactions with this modal."""

    value: str = attrs.field(repr=True)
    """Value provided for this text input."""


@attrs.define(kw_only=True, weakref_slot=False)
class MediaResource(files.Resource[files.AsyncReader]):
    """Represents a media resource."""

    resource: files.Resource[files.AsyncReader] = attrs.field(repr=True)
    """The resource this object wraps around."""

    proxy_resource: files.Resource[files.AsyncReader] | None = attrs.field(default=None, repr=False)
    """The proxied version of the resource, or [`None`][] if not present.

    !!! note
        This field cannot be set by bots or webhooks while sending an embed
        and will be ignored during serialization. Expect this to be
        populated on any received embed attached to a message event.
    """

    width: undefined.UndefinedNoneOr[int] = attrs.field(repr=True)
    """The width of media item."""

    height: undefined.UndefinedNoneOr[int] = attrs.field(repr=True)
    """The height of the media item."""

    content_type: undefined.UndefinedNoneOr[str] = attrs.field(repr=True)
    """The content type of the media item."""

    loading_state: undefined.UndefinedNoneOr[MediaLoadingType] = attrs.field(repr=True)
    """The loading state of the media item."""

    @property
    @typing_extensions.override
    def url(self) -> str:
        """URL of this embed resource."""
        return self.resource.url

    @property
    @typing_extensions.override
    def filename(self) -> str:
        """File name of this embed resource."""
        return self.resource.filename

    @property
    def proxy_url(self) -> str | None:
        """Proxied URL of this embed resource if applicable, else [`None`][]."""
        return self.proxy_resource.url if self.proxy_resource else None

    @property
    def proxy_filename(self) -> str | None:
        """File name of the proxied version of this embed resource if applicable, else [`None`][]."""
        return self.proxy_resource.filename if self.proxy_resource else None

    @typing_extensions.override
    def stream(
        self, *, executor: concurrent.futures.Executor | None = None, head_only: bool = False
    ) -> files.AsyncReaderContextManager[files.AsyncReader]:
        """Produce a stream of data for the resource.

        Parameters
        ----------
        executor
            The executor to run in for blocking operations.
            If [`None`][], then the default executor is used for the
            current event loop.
        head_only
            If [`True`][], then the implementation may only retrieve
            HEAD information if supported. This currently only has
            any effect for web requests.
        """
        return self.resource.stream(executor=executor, head_only=head_only)


@attrs.define(kw_only=True, weakref_slot=False)
class SectionComponent(PartialComponent):
    """Represents a section component."""

    components: typing.Sequence[SectionComponentTypesT] = attrs.field()
    """The sections components."""

    accessory: SectionAccessoryTypesT = attrs.field()
    """The sections accessory."""


@attrs.define(kw_only=True, weakref_slot=False)
class ThumbnailComponent(PartialComponent):
    """Represents a thumbnail component."""

    media: MediaResource = attrs.field()
    """The media for the thumbnail."""

    description: str | None = attrs.field()
    """The description of the thumbnail."""

    is_spoiler: bool = attrs.field()
    """Whether the thumbnail is marked as a spoiler."""


@attrs.define(kw_only=True, weakref_slot=False)
class TextDisplayComponent(PartialComponent):
    """Represents a text display component."""

    content: str = attrs.field()
    """The content of the text display."""


@attrs.define(kw_only=True, weakref_slot=False)
class MediaGalleryComponent(PartialComponent):
    """Represents a media gallery component."""

    items: typing.Sequence[MediaGalleryItem] = attrs.field()
    """The media gallery's items."""


@attrs.define(kw_only=True, weakref_slot=False)
class MediaGalleryItem:
    """Represents a media gallery item."""

    media: MediaResource = attrs.field()
    """The media for the gallery item."""

    description: str | None = attrs.field()
    """The description of the gallery item."""

    is_spoiler: bool = attrs.field()
    """Whether the gallery item is marked as a spoiler."""


@attrs.define(kw_only=True, weakref_slot=False)
class SeparatorComponent(PartialComponent):
    """Represents the separator component."""

    spacing: SpacingType = attrs.field()
    """The spacing for the separator."""

    divider: bool = attrs.field()
    """If there is a divider for the separator."""


@attrs.define(kw_only=True, weakref_slot=False)
class FileComponent(PartialComponent):
    """Represents a file component."""

    file: MediaResource = attrs.field()
    """The media for the file."""

    is_spoiler: bool = attrs.field()
    """If the file has a spoiler."""


@attrs.define(kw_only=True, weakref_slot=False)
class ContainerComponent(PartialComponent):
    """Represents a container component."""

    accent_color: colors.Color | None = attrs.field()
    """The accent colour for the container."""

    is_spoiler: bool = attrs.field()
    """Whether the container is marked as a spoiler."""

    components: typing.Sequence[ContainerTypesT] = attrs.field()
    """The components within the container."""


TopLevelComponentTypesT = typing.Union[
    ActionRowComponent[PartialComponent],
    TextDisplayComponent,
    SectionComponent,
    MediaGalleryComponent,
    SeparatorComponent,
    FileComponent,
    ContainerComponent,
]
"""Type hints of the values which are valid for top level components.

The following values are valid for this:

* [`hikari.components.ActionRowComponent`][]
* [`hikari.components.TextDisplayComponent`][]
* [`hikari.components.SectionComponent`][]
* [`hikari.components.MediaGalleryComponent`][]
* [`hikari.components.SeparatorComponent`][]
* [`hikari.components.FileComponent`][]
* [`hikari.components.ContainerComponent`][]
"""

ContainerTypesT = typing.Union[
    ActionRowComponent[PartialComponent],
    TextDisplayComponent,
    SectionComponent,
    MediaGalleryComponent,
    SeparatorComponent,
    FileComponent,
]
"""Type hints of the values which are valid for container components.

The following values are valid for this:

* [`hikari.components.ActionRowComponent`][]
* [`hikari.components.TextDisplayComponent`][]
* [`hikari.components.SectionComponent`][]
* [`hikari.components.MediaGalleryComponent`][]
* [`hikari.components.SeparatorComponent`][]
* [`hikari.components.FileComponent`][]
"""

SectionComponentTypesT = TextDisplayComponent
"""Type hints of the values which are valid for section components.

The following values are valid for this:

* [`hikari.components.TextDisplayComponent`][]
"""

SectionAccessoryTypesT = typing.Union[ButtonComponent, ThumbnailComponent]
"""Type hints of the values which are valid for section accessories.

The following values are valid for this:

* [`hikari.components.ButtonComponent`][]
* [`hikari.components.ThumbnailComponent`][]
"""

SelectMenuTypesT = typing.Literal[
    ComponentType.TEXT_SELECT_MENU,
    3,
    ComponentType.USER_SELECT_MENU,
    5,
    ComponentType.ROLE_SELECT_MENU,
    6,
    ComponentType.MENTIONABLE_SELECT_MENU,
    7,
    ComponentType.CHANNEL_SELECT_MENU,
    8,
]
"""Type hints of the [`hikari.components.ComponentType`][] values which are valid for select menus.

The following values are valid for this:

* [`hikari.components.ComponentType.TEXT_SELECT_MENU`][]/`3`
* [`hikari.components.ComponentType.USER_SELECT_MENU`][]/`5`
* [`hikari.components.ComponentType.ROLE_SELECT_MENU`][]/`6`
* [`hikari.components.ComponentType.MENTIONABLE_SELECT_MENU`][]`/`7`
* [`hikari.components.ComponentType.CHANNEL_SELECT_MENU`][]`/`8`
"""

SelectMenuTypes: typing.AbstractSet[SelectMenuTypesT] = frozenset(
    (
        ComponentType.TEXT_SELECT_MENU,
        ComponentType.USER_SELECT_MENU,
        ComponentType.ROLE_SELECT_MENU,
        ComponentType.MENTIONABLE_SELECT_MENU,
        ComponentType.CHANNEL_SELECT_MENU,
    )
)
"""Set of the [`hikari.components.ComponentType`][] values which are valid for select menus.

The following values are included in this:

* [`hikari.components.ComponentType.TEXT_SELECT_MENU`][]
* [`hikari.components.ComponentType.USER_SELECT_MENU`][]
* [`hikari.components.ComponentType.ROLE_SELECT_MENU`][]
* [`hikari.components.ComponentType.MENTIONABLE_SELECT_MENU`][]
* [`hikari.components.ComponentType.CHANNEL_SELECT_MENU`][]
"""

InteractiveButtonTypesT = typing.Literal[
    ButtonStyle.PRIMARY, 1, ButtonStyle.SECONDARY, 2, ButtonStyle.SUCCESS, 3, ButtonStyle.DANGER, 4
]
"""Type hints of the [`hikari.components.ButtonStyle`][] values which are valid for interactive buttons.

The following values are valid for this:

* [`hikari.components.ButtonStyle.PRIMARY`][]
* [`hikari.components.ButtonStyle.SECONDARY`][]
* [`hikari.components.ButtonStyle.SUCCESS`][]
* [`hikari.components.ButtonStyle.DANGER`][]
"""

InteractiveButtonTypes: typing.AbstractSet[InteractiveButtonTypesT] = frozenset(
    [ButtonStyle.PRIMARY, ButtonStyle.SECONDARY, ButtonStyle.SUCCESS, ButtonStyle.DANGER]
)
"""Set of the [`hikari.components.ButtonStyle`][] which are valid for interactive buttons.

The following values are included in this:

* [`hikari.components.ButtonStyle.PRIMARY`][]
* [`hikari.components.ButtonStyle.SECONDARY`][]
* [`hikari.components.ButtonStyle.SUCCESS`][]
* [`hikari.components.ButtonStyle.DANGER`][]
"""

MessageComponentTypesT = typing.Union[ButtonComponent, SelectMenuComponent]
"""Type hint of the [`hikari.components.PartialComponent`][] that be contained in a [`hikari.components.PartialComponent`][].

The following values are valid for this:

* [`hikari.components.ButtonComponent`][]
* [`hikari.components.SelectMenuComponent`][]
"""  # noqa: E501
ModalComponentTypesT = TextInputComponent
"""Type hint of the [`hikari.components.PartialComponent`][] that be contained in a [`hikari.components.PartialComponent`][].

The following values are valid for this:

* [`hikari.components.TextInputComponent`][]
"""  # noqa: E501

MessageActionRowComponent = ActionRowComponent[MessageComponentTypesT]
"""A message action row component."""
ModalActionRowComponent = ActionRowComponent[ModalComponentTypesT]
"""A modal action row component."""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import templates


class TestTemplate:
    @pytest.fixture
    def obj(self):
        return templates.Template(
            app=mock.Mock(),
            code="abc123",
            name="Test Template",
            description="Template used for testing",
            usage_count=101,
            creator=object(),
            created_at=object(),
            updated_at=object(),
            source_guild=mock.Mock(id=123),
            is_unsynced=True,
        )

    @pytest.mark.asyncio
    async def test_fetch_self(self, obj):
        obj.app.rest.fetch_template = mock.AsyncMock()

        assert await obj.fetch_self() is obj.app.rest.fetch_template.return_value

        obj.app.rest.fetch_template.assert_awaited_once_with("abc123")

    @pytest.mark.asyncio
    async def test_edit(self, obj):
        obj.app.rest.edit_template = mock.AsyncMock()

        returned = await obj.edit(name="Test Template 2", description="Electric Boogaloo")
        assert returned is obj.app.rest.edit_template.return_value

        obj.app.rest.edit_template.assert_awaited_once_with(
            obj.source_guild, obj, name="Test Template 2", description="Electric Boogaloo"
        )

    @pytest.mark.asyncio
    async def test_delete(self, obj):
        obj.app.rest.delete_template = mock.AsyncMock()

        await obj.delete()

        obj.app.rest.delete_template.assert_awaited_once_with(obj.source_guild, obj)

    @pytest.mark.asyncio
    async def test_sync(self, obj):
        obj.app.rest.sync_guild_template = mock.AsyncMock()

        assert await obj.sync() is obj.app.rest.sync_guild_template.return_value

        obj.app.rest.sync_guild_template.assert_awaited_once_with(123, "abc123")

    def test_str(self, obj):
        assert str(obj) == "https://discord.new/abc123"



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import applications
from hikari import monetization
from hikari import snowflakes
from hikari import traits
from hikari import undefined
from hikari.interactions import base_interactions


@pytest.fixture
def mock_app():
    return mock.Mock(traits.CacheAware, rest=mock.AsyncMock())


class TestPartialInteraction:
    @pytest.fixture
    def mock_partial_interaction(self, mock_app):
        return base_interactions.PartialInteraction(
            app=mock_app,
            id=34123,
            application_id=651231,
            type=base_interactions.InteractionType.APPLICATION_COMMAND,
            token="399393939doodsodso",
            version=3122312,
            guild_id=snowflakes.Snowflake(5412231),
            channel=mock.Mock(id=3123123),
            member=object(),
            user=object(),
            locale="es-ES",
            guild_locale="en-US",
            app_permissions=123321,
            authorizing_integration_owners={
                applications.ApplicationIntegrationType.GUILD_INSTALL: snowflakes.Snowflake(123)
            },
            entitlements=[
                monetization.Entitlement(
                    id=snowflakes.Snowflake(123123),
                    sku_id=snowflakes.Snowflake(123123),
                    application_id=snowflakes.Snowflake(123123),
                    guild_id=snowflakes.Snowflake(123123),
                    user_id=snowflakes.Snowflake(123123),
                    type=monetization.EntitlementType.APPLICATION_SUBSCRIPTION,
                    starts_at=None,
                    ends_at=None,
                    is_deleted=False,
                    subscription_id=None,
                )
            ],
            context=applications.ApplicationContextType.PRIVATE_CHANNEL,
        )

    def test_webhook_id_property(self, mock_partial_interaction):
        assert mock_partial_interaction.webhook_id is mock_partial_interaction.application_id

    @pytest.mark.asyncio
    async def test_fetch_guild(self, mock_partial_interaction, mock_app):
        mock_partial_interaction.guild_id = 43123123

        assert await mock_partial_interaction.fetch_guild() is mock_app.rest.fetch_guild.return_value

        mock_app.rest.fetch_guild.assert_awaited_once_with(43123123)

    @pytest.mark.asyncio
    async def test_fetch_guild_for_dm_interaction(self, mock_partial_interaction, mock_app):
        mock_partial_interaction.guild_id = None

        assert await mock_partial_interaction.fetch_guild() is None

        mock_app.rest.fetch_guild.assert_not_called()

    def test_get_guild(self, mock_partial_interaction, mock_app):
        mock_partial_interaction.guild_id = 874356

        assert mock_partial_interaction.get_guild() is mock_app.cache.get_guild.return_value

        mock_app.cache.get_guild.assert_called_once_with(874356)

    def test_get_guild_for_dm_interaction(self, mock_partial_interaction, mock_app):
        mock_partial_interaction.guild_id = None

        assert mock_partial_interaction.get_guild() is None

        mock_app.cache.get_guild.assert_not_called()

    def test_get_guild_when_cacheless(self, mock_partial_interaction, mock_app):
        mock_partial_interaction.guild_id = 321123
        mock_partial_interaction.app = mock.Mock(traits.RESTAware)

        assert mock_partial_interaction.get_guild() is None

        mock_app.cache.get_guild.assert_not_called()


class TestMessageResponseMixin:
    @pytest.fixture
    def mock_message_response_mixin(self, mock_app):
        return base_interactions.MessageResponseMixin(
            app=mock_app,
            id=34123,
            application_id=651231,
            type=base_interactions.InteractionType.APPLICATION_COMMAND,
            token="399393939doodsodso",
            version=3122312,
            context=applications.ApplicationContextType.PRIVATE_CHANNEL,
            guild_id=snowflakes.Snowflake(5412231),
            channel=mock.Mock(id=3123123),
            member=object(),
            user=object(),
            locale="es-ES",
            guild_locale="en-US",
            app_permissions=123321,
            authorizing_integration_owners={
                applications.ApplicationIntegrationType.GUILD_INSTALL: snowflakes.Snowflake(123)
            },
            entitlements=[
                monetization.Entitlement(
                    id=snowflakes.Snowflake(123123),
                    sku_id=snowflakes.Snowflake(123123),
                    application_id=snowflakes.Snowflake(123123),
                    guild_id=snowflakes.Snowflake(123123),
                    user_id=snowflakes.Snowflake(123123),
                    type=monetization.EntitlementType.APPLICATION_SUBSCRIPTION,
                    starts_at=None,
                    ends_at=None,
                    is_deleted=False,
                    subscription_id=None,
                )
            ],
        )

    @pytest.mark.asyncio
    async def test_fetch_initial_response(self, mock_message_response_mixin, mock_app):
        result = await mock_message_response_mixin.fetch_initial_response()

        assert result is mock_app.rest.fetch_interaction_response.return_value
        mock_app.rest.fetch_interaction_response.assert_awaited_once_with(651231, "399393939doodsodso")

    @pytest.mark.asyncio
    async def test_create_initial_response_with_optional_args(self, mock_message_response_mixin, mock_app):
        mock_embed_1 = object()
        mock_embed_2 = object()
        mock_poll = object()
        mock_component = object()
        mock_components = object(), object()
        mock_attachment = object()
        mock_attachments = object(), object()
        await mock_message_response_mixin.create_initial_response(
            base_interactions.ResponseType.MESSAGE_CREATE,
            "content",
            tts=True,
            embed=mock_embed_1,
            flags=64,
            embeds=[mock_embed_2],
            poll=mock_poll,
            component=mock_component,
            components=mock_components,
            attachment=mock_attachment,
            attachments=mock_attachments,
            mentions_everyone=False,
            user_mentions=[123432],
            role_mentions=[6324523],
        )

        mock_app.rest.create_interaction_response.assert_awaited_once_with(
            34123,
            "399393939doodsodso",
            base_interactions.ResponseType.MESSAGE_CREATE,
            "content",
            tts=True,
            flags=64,
            embed=mock_embed_1,
            embeds=[mock_embed_2],
            poll=mock_poll,
            component=mock_component,
            components=mock_components,
            attachment=mock_attachment,
            attachments=mock_attachments,
            mentions_everyone=False,
            user_mentions=[123432],
            role_mentions=[6324523],
        )

    @pytest.mark.asyncio
    async def test_create_initial_response_without_optional_args(self, mock_message_response_mixin, mock_app):
        await mock_message_response_mixin.create_initial_response(
            base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE
        )

        mock_app.rest.create_interaction_response.assert_awaited_once_with(
            34123,
            "399393939doodsodso",
            base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE,
            undefined.UNDEFINED,
            flags=undefined.UNDEFINED,
            tts=undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            poll=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
        )

    @pytest.mark.asyncio
    async def test_edit_initial_response_with_optional_args(self, mock_message_response_mixin, mock_app):
        mock_embed_1 = object()
        mock_embed_2 = object()
        mock_attachment_1 = object()
        mock_attachment_2 = object()
        mock_component = object()
        mock_components = object(), object()
        result = await mock_message_response_mixin.edit_initial_response(
            "new content",
            embed=mock_embed_1,
            embeds=[mock_embed_2],
            attachment=mock_attachment_1,
            attachments=[mock_attachment_2],
            component=mock_component,
            components=mock_components,
            mentions_everyone=False,
            user_mentions=[123123],
            role_mentions=[562134],
        )

        assert result is mock_app.rest.edit_interaction_response.return_value
        mock_app.rest.edit_interaction_response.assert_awaited_once_with(
            651231,
            "399393939doodsodso",
            "new content",
            embed=mock_embed_1,
            embeds=[mock_embed_2],
            attachment=mock_attachment_1,
            attachments=[mock_attachment_2],
            component=mock_component,
            components=mock_components,
            mentions_everyone=False,
            user_mentions=[123123],
            role_mentions=[562134],
        )

    @pytest.mark.asyncio
    async def test_edit_initial_response_without_optional_args(self, mock_message_response_mixin, mock_app):
        result = await mock_message_response_mixin.edit_initial_response()

        assert result is mock_app.rest.edit_interaction_response.return_value
        mock_app.rest.edit_interaction_response.assert_awaited_once_with(
            651231,
            "399393939doodsodso",
            undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
        )

    @pytest.mark.asyncio
    async def test_delete_initial_response(self, mock_message_response_mixin, mock_app):
        await mock_message_response_mixin.delete_initial_response()

        mock_app.rest.delete_interaction_response.assert_awaited_once_with(651231, "399393939doodsodso")


class TestModalResponseMixin:
    @pytest.fixture
    def mock_modal_response_mixin(self, mock_app):
        return base_interactions.ModalResponseMixin(
            app=mock_app,
            id=34123,
            application_id=651231,
            type=base_interactions.InteractionType.APPLICATION_COMMAND,
            token="399393939doodsodso",
            version=3122312,
            context=applications.ApplicationContextType.PRIVATE_CHANNEL,
            guild_id=snowflakes.Snowflake(5412231),
            channel=mock.Mock(id=3123123),
            member=object(),
            user=object(),
            locale="es-ES",
            guild_locale="en-US",
            app_permissions=123321,
            authorizing_integration_owners={
                applications.ApplicationIntegrationType.GUILD_INSTALL: snowflakes.Snowflake(123)
            },
            entitlements=[
                monetization.Entitlement(
                    id=snowflakes.Snowflake(123123),
                    sku_id=snowflakes.Snowflake(123123),
                    application_id=snowflakes.Snowflake(123123),
                    guild_id=snowflakes.Snowflake(123123),
                    user_id=snowflakes.Snowflake(123123),
                    type=monetization.EntitlementType.APPLICATION_SUBSCRIPTION,
                    starts_at=None,
                    ends_at=None,
                    is_deleted=False,
                    subscription_id=None,
                )
            ],
        )

    @pytest.mark.asyncio
    async def test_create_modal_response(self, mock_modal_response_mixin, mock_app):
        await mock_modal_response_mixin.create_modal_response("title", "custom_id", None, [])

        mock_app.rest.create_modal_response.assert_awaited_once_with(
            34123, "399393939doodsodso", title="title", custom_id="custom_id", component=None, components=[]
        )

    def test_build_response(self, mock_modal_response_mixin, mock_app):
        mock_app.rest.interaction_modal_builder = mock.Mock()
        builder = mock_modal_response_mixin.build_modal_response("title", "custom_id")

        assert builder is mock_app.rest.interaction_modal_builder.return_value
        mock_app.rest.interaction_modal_builder.assert_called_once_with(title="title", custom_id="custom_id")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Application and entities that are used to describe guild scheduled events on Discord."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "EventPrivacyLevel",
    "ScheduledEvent",
    "ScheduledEventStatus",
    "ScheduledEventType",
    "ScheduledEventUser",
    "ScheduledExternalEvent",
    "ScheduledStageEvent",
    "ScheduledVoiceEvent",
)

import typing

import attrs

from hikari import snowflakes
from hikari import undefined
from hikari import urls
from hikari.internal import attrs_extensions
from hikari.internal import deprecation
from hikari.internal import enums
from hikari.internal import routes

if typing.TYPE_CHECKING:
    import datetime

    from hikari import files
    from hikari import guilds
    from hikari import traits
    from hikari import users


class EventPrivacyLevel(int, enums.Enum):
    """Enum of the possible scheduled event privacy levels."""

    GUILD_ONLY = 2
    """The scheduled event is only available to guild members."""


class ScheduledEventType(int, enums.Enum):
    """Enum of the scheduled event types."""

    STAGE_INSTANCE = 1
    """A scheduled stage instance."""

    VOICE = 2
    """A scheduled voice chat event."""

    EXTERNAL = 3
    """A scheduled event which takes part outside of Discord."""


class ScheduledEventStatus(int, enums.Enum):
    """Enum of the scheduled event statuses."""

    SCHEDULED = 1
    """Indicates that the scheduled event hasn't occurred yet."""

    ACTIVE = 2
    """Indicates an event is on-going."""

    COMPLETED = 3
    """Indicates an event has finished."""

    CANCELED = 4
    """Indicates an event has been canceled."""

    CANCELLED = CANCELED
    """Alias of [`hikari.scheduled_events.ScheduledEventStatus.CANCELED`][]."""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ScheduledEvent(snowflakes.Unique):
    """Base class for scheduled events."""

    # entity_id is ignored right now due to always being null
    # creator_id is ignored as it just dupes creator.id

    app: traits.RESTAware = attrs.field(
        repr=False, eq=False, hash=False, metadata={attrs_extensions.SKIP_DEEP_COPY: True}
    )
    """Client application that models may use for procedures."""

    id: snowflakes.Snowflake = attrs.field(hash=True, repr=True)
    """ID of the scheduled event."""

    guild_id: snowflakes.Snowflake = attrs.field(hash=False, repr=True)
    """ID of the guild this scheduled event belongs to."""

    name: str = attrs.field(hash=False, repr=True)
    """Name of the scheduled event."""

    description: str | None = attrs.field(hash=False, repr=False)
    """Description of the scheduled event."""

    start_time: datetime.datetime = attrs.field(hash=False, repr=False)
    """When the event is scheduled to start."""

    end_time: datetime.datetime | None = attrs.field(hash=False, repr=False)
    """When the event is scheduled to end, if set."""

    privacy_level: EventPrivacyLevel = attrs.field(hash=False, repr=False)
    """Privacy level of the scheduled event.

    This restricts who can view and join the scheduled event.
    """

    status: ScheduledEventStatus = attrs.field(hash=False, repr=True)
    """Status of the scheduled event."""

    entity_type: ScheduledEventType = attrs.field(hash=False, repr=True)
    """The type of entity this scheduled event is associated with."""

    creator: users.User | None = attrs.field(hash=False, repr=False)
    """The user who created the scheduled event.

    This will only be set for event created after 2021-10-25.
    """

    user_count: int | None = attrs.field(hash=False, repr=False)
    """The number of users that have subscribed to the event.

    This will be [`None`][] on gateway events when creating and
    editing a scheduled event.
    """

    image_hash: str | None = attrs.field(hash=False, repr=False)
    """Hash of the image used for the scheduled event, if set."""

    @property
    @deprecation.deprecated("Use 'make_image_url' instead.")
    def image_url(self) -> files.URL | None:
        """Cover image for this scheduled event, if set."""
        deprecation.warn_deprecated(
            "image_url", removal_version="2.5.0", additional_info="Use 'make_image_url' instead."
        )
        return self.make_image_url()

    def make_image_url(
        self,
        *,
        file_format: typing.Literal["PNG", "JPEG", "JPG", "WEBP"] = "PNG",
        size: int = 4096,
        lossless: bool = True,
        ext: str | None | undefined.UndefinedType = undefined.UNDEFINED,
    ) -> files.URL | None:
        """Generate the cover image URL for this scheduled event, if set.

        If no cover image is set, this returns [`None`][].

        Parameters
        ----------
        file_format
            The format to use for this URL.

            Supports `PNG`, `JPEG`, `JPG`, and `WEBP`.

            If not specified, the format will be `PNG`.
        size
            The size to set for the URL;
            Can be any power of two between `16` and `4096`;
        lossless
            Whether to return a lossless or compressed WEBP image;
            This is ignored if `file_format` is not `WEBP`.
        ext
            The extension to use for this URL.
            Supports `png`, `jpeg`, `jpg` and `webp`.

            !!! deprecated 2.4.0
                This has been replaced with the `file_format` argument.

        Returns
        -------
        typing.Optional[hikari.files.URL]
            The URL, or [`None`][] if no cover image is set.

        Raises
        ------
        TypeError
            If an invalid format is passed for `file_format`.
        ValueError
            If `size` is specified but is not a power of two or not between 16 and 4096.
        """
        if self.image_hash is None:
            return None

        if ext:
            deprecation.warn_deprecated(
                "ext", removal_version="2.5.0", additional_info="Use 'file_format' argument instead."
            )
            file_format = ext.upper()  # type: ignore[assignment]

        return routes.SCHEDULED_EVENT_COVER.compile_to_file(
            urls.CDN_URL,
            scheduled_event_id=self.id,
            hash=self.image_hash,
            size=size,
            file_format=file_format,
            lossless=lossless,
        )


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ScheduledExternalEvent(ScheduledEvent):
    """A scheduled event that takes place outside of Discord."""

    location: str = attrs.field(hash=False, repr=False)
    """The location of the scheduled event.

    !!! note
        There is no strict format for this field, and it will likely be a user
        friendly string.
    """

    end_time: datetime.datetime = attrs.field(hash=False, repr=False)
    """When the event is scheduled to end."""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ScheduledStageEvent(ScheduledEvent):
    """A scheduled event that takes place in a stage channel."""

    channel_id: snowflakes.Snowflake = attrs.field(hash=False, repr=False)
    """ID of the stage channel this event is scheduled in."""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ScheduledVoiceEvent(ScheduledEvent):
    """A scheduled event that takes place in a voice channel."""

    channel_id: snowflakes.Snowflake = attrs.field(hash=False, repr=False)
    """ID of the voice channel this scheduled event is in."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ScheduledEventUser:
    """A user who is subscribed to a scheduled event."""

    event_id: snowflakes.Snowflake = attrs.field(hash=False, repr=True)
    """ID of the scheduled event they're subscribed to."""

    user: users.User = attrs.field(hash=True, repr=True)
    """Object representing the user."""

    member: guilds.Member | None = attrs.field(hash=False, repr=False)
    """Their guild member object if they're in the event's guild."""



# Hikari Examples - A collection of examples for Hikari.
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""An example OAuth server."""

from __future__ import annotations

import logging
import os

from aiohttp import web

import hikari

logging.basicConfig(level=logging.DEBUG)

host = "localhost"
port = 8080
CLIENT_ID = int(os.environ["CLIENT_ID"])  # ID as an int
CLIENT_SECRET = os.environ["CLIENT_SECRET"]  # Secret as a str
BOT_TOKEN = os.environ["BOT_TOKEN"]  # Token as a str
CHANNEL_ID = int(os.environ["CHANNEL_ID"])  # Channel to post in as an int
REDIRECT_URI = "http://localhost:8080"

route_table = web.RouteTableDef()


@route_table.get("/")
async def oauth(request: web.Request) -> web.Response:
    """Handle an OAuth request."""
    code = request.query.get("code")
    if not code:
        return web.json_response({"error": "'code' is not provided"}, status=400)

    discord_rest: hikari.RESTApp = request.app["discord_rest"]

    # Exchange code to acquire a Bearer one for the user
    async with discord_rest.acquire(None) as r:
        auth = await r.authorize_access_token(CLIENT_ID, CLIENT_SECRET, code, REDIRECT_URI)

    # Perform a request as the user to get their own user object
    async with discord_rest.acquire(auth.access_token, hikari.TokenType.BEARER) as client:
        user = await client.fetch_my_user()
        # user is a hikari.OwnUser object where we can access attributes on it

    # Notify the success
    async with discord_rest.acquire(BOT_TOKEN, hikari.TokenType.BOT) as client:
        await client.create_message(CHANNEL_ID, f"{user} ({user.id}) just authorized!")

    return web.Response(text="Successfully authenticated!")


async def start_discord_rest(app: web.Application) -> None:
    """Start the RESTApp."""
    discord_rest = hikari.RESTApp()
    await discord_rest.start()

    app["discord_rest"] = discord_rest


async def stop_discord_rest(app: web.Application) -> None:
    """Stop the RESTApp."""
    discord_rest: hikari.RESTApp = app["discord_rest"]

    await discord_rest.close()


if __name__ == "__main__":
    server = web.Application()
    server.add_routes(route_table)

    server.on_startup.append(start_discord_rest)
    server.on_cleanup.append(stop_discord_rest)

    web.run_app(server, host=host, port=port)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime

import mock
import pytest

from hikari.internal import time


def test_parse_iso_8601_date_with_negative_timezone():
    string = "2019-10-10T05:22:33.023456-02:30"
    date = time.iso8601_datetime_string_to_datetime(string)
    assert date.year == 2019
    assert date.month == 10
    assert date.day == 10
    assert date.hour == 5
    assert date.minute == 22
    assert date.second == 33
    assert date.microsecond == 23456
    offset = date.tzinfo.utcoffset(None)
    assert offset == datetime.timedelta(hours=-2, minutes=-30)


def test_slow_parse_iso_8601_date_with_positive_timezone():
    string = "2019-10-10T05:22:33.023456+02:30"
    date = time.iso8601_datetime_string_to_datetime(string)
    assert date.year == 2019
    assert date.month == 10
    assert date.day == 10
    assert date.hour == 5
    assert date.minute == 22
    assert date.second == 33
    assert date.microsecond == 23456
    offset = date.tzinfo.utcoffset(None)
    assert offset == datetime.timedelta(hours=2, minutes=30)


def test_parse_iso_8601_date_with_zulu():
    string = "2019-10-10T05:22:33.023456Z"
    date = time.iso8601_datetime_string_to_datetime(string)
    assert date.year == 2019
    assert date.month == 10
    assert date.day == 10
    assert date.hour == 5
    assert date.minute == 22
    assert date.second == 33
    assert date.microsecond == 23456
    offset = date.tzinfo.utcoffset(None)
    assert offset == datetime.timedelta(seconds=0)


def test_parse_iso_8601_date_with_milliseconds_instead_of_microseconds():
    string = "2019-10-10T05:22:33.023Z"
    date = time.iso8601_datetime_string_to_datetime(string)
    assert date.year == 2019
    assert date.month == 10
    assert date.day == 10
    assert date.hour == 5
    assert date.minute == 22
    assert date.second == 33
    assert date.microsecond == 23000


def test_parse_iso_8601_date_with_no_fraction():
    string = "2019-10-10T05:22:33Z"
    date = time.iso8601_datetime_string_to_datetime(string)
    assert date.year == 2019
    assert date.month == 10
    assert date.day == 10
    assert date.hour == 5
    assert date.minute == 22
    assert date.second == 33
    assert date.microsecond == 0


def test_speedup_replaces_python_version_when_available():
    try:
        import ciso8601

        assert time.fast_iso8601_datetime_string_to_datetime is ciso8601.parse_rfc3339
        assert time.iso8601_datetime_string_to_datetime is ciso8601.parse_rfc3339
    except ModuleNotFoundError:
        # No speedups, test pure version has been setup correctly

        assert time.fast_iso8601_datetime_string_to_datetime is None
        assert time.iso8601_datetime_string_to_datetime is time.slow_iso8601_datetime_string_to_datetime


def test_parse_discord_epoch_to_datetime():
    discord_timestamp = 37921278956
    expected_timestamp = datetime.datetime(2016, 3, 14, 21, 41, 18, 956000, tzinfo=datetime.timezone.utc)
    assert time.discord_epoch_to_datetime(discord_timestamp) == expected_timestamp


def test_parse_datetime_to_discord_epoch():
    # This specific timestamp (amongst others) has given problems in the past with
    # float precision, so make sure it doesn't happen again.
    timestamp = datetime.datetime(2022, 8, 13, 23, 10, 35, 843000, tzinfo=datetime.timezone.utc)
    expected_discord_timestamp = 240361835843
    assert time.datetime_to_discord_epoch(timestamp) == expected_discord_timestamp


def test_parse_unix_epoch_to_datetime():
    unix_timestamp = 1457991678956
    expected_timestamp = datetime.datetime(2016, 3, 14, 21, 41, 18, 956000, tzinfo=datetime.timezone.utc)
    assert time.unix_epoch_to_datetime(unix_timestamp) == expected_timestamp


def test_unix_epoch_to_datetime_with_out_of_range_positive_timestamp():
    assert time.unix_epoch_to_datetime(996877846784536) == datetime.datetime.max.replace(tzinfo=datetime.timezone.utc)


def test_unix_epoch_to_datetime_with_out_of_range_negative_timestamp():
    assert time.unix_epoch_to_datetime(-996877846784536) == datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)


@pytest.mark.parametrize(
    ("input_value", "expected_result"),
    [
        (5, 5),
        (2.718281828459045, 2),
        (datetime.timedelta(days=5, seconds=3, milliseconds=12), 432_003),
        (-5, 0),
        (-2.718281828459045, 0),
        (datetime.timedelta(days=-5, seconds=-3, milliseconds=12), 0),
    ],
)
def test_timespan_to_int(input_value, expected_result):
    assert time.timespan_to_int(input_value) == expected_result


def test_utc_datetime():
    current_datetime = datetime.datetime.now(tz=datetime.timezone.utc)

    # We can't mock datetime normally as it is a C module :(
    class datetime_module:
        timezone = datetime.timezone

        class datetime:
            now = mock.Mock(return_value=current_datetime)

    with mock.patch.object(time, "datetime", datetime_module):
        result = time.utc_datetime()

    datetime_module.datetime.now.assert_called_once_with(tz=datetime.timezone.utc)

    assert result == current_datetime


def test_local_datetime():
    current_datetime = datetime.datetime.now(tz=datetime.timezone.utc)

    # We can't mock datetime normally as it is a C module :(
    class datetime_module:
        timezone = datetime.timezone

        class datetime:
            now = mock.Mock(return_value=current_datetime)

    with mock.patch.object(time, "datetime", datetime_module):
        result = time.local_datetime()

    datetime_module.datetime.now.assert_called_once_with(tz=datetime.timezone.utc)

    assert result == current_datetime.astimezone()



# Hikari Examples - A collection of examples for Hikari.
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""A simple bot with some simple commands."""

from __future__ import annotations

import os

import hikari

bot = hikari.GatewayBot(token=os.environ["BOT_TOKEN"])

# Set this to the guild to create the slash commands in.
# To create global commands, set to `hikari.UNDEFINED`.
COMMAND_GUILD_ID = 0


@bot.listen()
async def register_commands(_: hikari.StartingEvent) -> None:
    """Register ping and info commands."""
    application = await bot.rest.fetch_application()

    commands = [
        bot.rest.slash_command_builder("ping", "Get the bot's latency."),
        bot.rest.slash_command_builder("info", "Learn something about the bot."),
        bot.rest.slash_command_builder("ephemeral", "Send a very secret message."),
    ]

    await bot.rest.set_application_commands(application=application.id, commands=commands, guild=COMMAND_GUILD_ID)


@bot.listen()
async def handle_interactions(event: hikari.CommandInteractionCreateEvent) -> None:
    """Listen for slash commands being executed."""
    if event.interaction.command_name == "ping":
        await event.interaction.create_initial_response(
            hikari.ResponseType.MESSAGE_CREATE, f"Pong! {bot.heartbeat_latency * 1_000:.0f}ms"
        )

    elif event.interaction.command_name == "info":
        await event.interaction.create_initial_response(
            hikari.ResponseType.MESSAGE_CREATE, "Hello, this is an example bot written in hikari!"
        )

    elif event.interaction.command_name == "ephemeral":
        await event.interaction.create_initial_response(
            hikari.ResponseType.MESSAGE_CREATE,
            "Only you can see this, keep it a secret :)",
            flags=hikari.MessageFlag.EPHEMERAL,
        )


bot.run()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import guilds
from hikari import snowflakes
from hikari.events import member_events
from tests.hikari import hikari_test_helpers


class TestMemberEvent:
    @pytest.fixture
    def event(self):
        cls = hikari_test_helpers.mock_class_namespace(
            member_events.MemberEvent,
            slots_=False,
            guild_id=mock.PropertyMock(return_value=snowflakes.Snowflake(123)),
            user=mock.Mock(id=456),
        )
        return cls()

    def test_app_property(self, event):
        assert event.app is event.user.app

    def test_user_id_property(self, event):
        event.user_id == 456

    def test_guild_when_no_cache_trait(self):
        event = hikari_test_helpers.mock_class_namespace(member_events.MemberEvent, app=None)()

        assert event.get_guild() is None

    def test_get_guild_when_available(self, event):
        result = event.get_guild()

        assert result is event.app.cache.get_available_guild.return_value
        event.app.cache.get_available_guild.assert_called_once_with(123)
        event.app.cache.get_unavailable_guild.assert_not_called()

    def test_guild_when_unavailable(self, event):
        event.app.cache.get_available_guild.return_value = None
        result = event.get_guild()

        assert result is event.app.cache.get_unavailable_guild.return_value
        event.app.cache.get_unavailable_guild.assert_called_once_with(123)
        event.app.cache.get_available_guild.assert_called_once_with(123)


class TestMemberCreateEvent:
    @pytest.fixture
    def event(self):
        return member_events.MemberCreateEvent(shard=None, member=mock.Mock())

    def test_guild_property(self, event):
        event.member.guild_id = 123
        event.guild_id == 123

    def test_user_property(self, event):
        user = object()
        event.member.user = user
        event.user == user


class TestMemberUpdateEvent:
    @pytest.fixture
    def event(self):
        return member_events.MemberUpdateEvent(shard=None, member=mock.Mock(), old_member=mock.Mock(guilds.Member))

    def test_guild_property(self, event):
        event.member.guild_id = 123
        event.guild_id == 123

    def test_user_property(self, event):
        user = object()
        event.member.user = user
        event.user == user

    def test_old_user_property(self, event):
        event.member.guild_id = 123
        event.member.id = 456

        assert event.member.guild_id == 123
        assert event.member.id == 456



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import contextlib
import typing

import mock
import pytest

from hikari.internal import fast_protocol


class TestCheckIfIgnored:
    def test_when_startswith_abc(self):
        mock_string = mock.Mock(startswith=mock.Mock(return_value=True))

        assert fast_protocol._check_if_ignored(mock_string) is True

        mock_string.startswith.assert_called_once_with("_abc_")

    def test_when_in_IGNORED_ATTRS(self):
        with mock.patch.object(
            fast_protocol, "_IGNORED_ATTRS", new=mock.Mock(__contains__=mock.Mock(return_value=True))
        ) as mock_list:
            assert fast_protocol._check_if_ignored("testing") is True

            mock_list.__contains__.assert_called_once_with("testing")

    def test_when_not_ignored(self):
        mock_string = mock.Mock(startswith=mock.Mock(return_value=False))
        with mock.patch.object(
            fast_protocol, "_IGNORED_ATTRS", new=mock.Mock(__contains__=mock.Mock(return_value=False))
        ) as mock_list:
            assert fast_protocol._check_if_ignored(mock_string) is False

            mock_list.__contains__.assert_called_once_with(mock_string)


class TestFastProtocolChecking:
    def test_new_when_first_protocol_not_FastProtocolChecking(self):
        stack = contextlib.ExitStack()
        stack.enter_context(
            pytest.raises(TypeError, match=r"First instance of _FastProtocolChecking must be FastProtocolChecking")
        )
        stack.enter_context(mock.patch.object(fast_protocol, "_Protocol", new=NotImplemented))

        with stack:

            class NotProtocol(metaclass=fast_protocol._FastProtocolChecking): ...

    def test_new_when_first_protocol_is_FastProtocolChecking(self):
        with mock.patch.object(fast_protocol, "_Protocol", new=NotImplemented):

            class FastProtocolChecking(metaclass=fast_protocol._FastProtocolChecking): ...

            assert fast_protocol._Protocol is FastProtocolChecking

        assert FastProtocolChecking._attributes_ == ()

    def test_new_when_bases_not_fastprotocols(self):
        with pytest.raises(
            TypeError,
            match=r"FastProtocolChecking can only inherit from other fast checking protocols, got <class 'str'>",
        ):

            class MyProtocol(str, fast_protocol.FastProtocolChecking, typing.Protocol): ...

    def test_new_when_fastprotocolchecking_in_bases_but_not_protocol(self):
        with pytest.raises(TypeError, match=r"FastProtocolChecking can only be used with protocols"):

            class MyProtocol(fast_protocol.FastProtocolChecking): ...

    def test_new(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
            def test1(): ...

        class OtherProtocol(MyProtocol, fast_protocol.FastProtocolChecking, typing.Protocol):
            def test2(): ...

        assert sorted(OtherProtocol._attributes_) == ["test1", "test2"]

    def test_init_subclass_does_not_overwrite_subclasshoook(self):
        def subclass_hook(): ...

        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
            __subclasshook__ = subclass_hook

        assert MyProtocol.__subclasshook__ is subclass_hook

    def test_isinstance_when_not_protocol(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol): ...

        class Class: ...

        MyProtocol._is_protocol = False
        class_instance = Class()

        with mock.patch.object(type(typing.Protocol), "__instancecheck__", return_value=True) as instance_check:
            assert isinstance(class_instance, MyProtocol) is True

            instance_check.assert_called_once_with(class_instance)

    def test_isinstance_fastfail(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
            def test(): ...

        class Class: ...

        assert isinstance(Class(), MyProtocol) is False

    def test_isinstance(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
            def test(): ...

        class Class:
            def test(): ...

        assert isinstance(Class(), MyProtocol) is True

    def test_issubclass_when_not_protocol(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol): ...

        class Class: ...

        with mock.patch.object(
            fast_protocol._FastProtocolChecking, "__subclasscheck__", return_value=True
        ) as subclass_check:
            assert issubclass(Class, MyProtocol) is True

            subclass_check.assert_called_once_with(Class)

    def test_issubclass_fastfail(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
            def test(): ...

        class Class: ...

        assert issubclass(Class, MyProtocol) is False

    def test_issubclass(self):
        class MyProtocol(fast_protocol.FastProtocolChecking, typing.Protocol):
            def test(): ...

        class Class:
            def test(): ...

        assert issubclass(Class, MyProtocol) is True



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import contextlib
import copy as stdlib_copy

import attrs
import mock

from hikari.internal import attrs_extensions


def test_invalidate_shallow_copy_cache():
    attrs_extensions._SHALLOW_COPIERS = {int: object(), str: object()}
    assert attrs_extensions.invalidate_shallow_copy_cache() is None
    assert attrs_extensions._SHALLOW_COPIERS == {}


def test_invalidate_deep_copy_cache():
    attrs_extensions._DEEP_COPIERS = {str: object(), int: object(), object: object()}
    assert attrs_extensions.invalidate_deep_copy_cache() is None
    assert attrs_extensions._DEEP_COPIERS == {}


def test_get_fields_definition():
    @attrs.define()
    class StubModel:
        foo: int = attrs.field()
        bar: bool = attrs.field(init=False)
        bam: bool = attrs.field(init=False)
        _voodoo: str = attrs.field(alias="voodoo")
        Bat: bool = attrs.field()

    fields = {field.name: field for field in attrs.fields(StubModel)}
    new_model = attrs_extensions.get_fields_definition(StubModel)
    assert new_model == (
        [(fields["foo"], "foo"), (fields["_voodoo"], "voodoo"), (fields["Bat"], "Bat")],
        [fields["bar"], fields["bam"]],
    )


def test_generate_shallow_copier():
    @attrs.define()
    class StubModel:
        _foo: int = attrs.field(alias="foo")
        baaaa: str = attrs.field()
        _blam: bool = attrs.field(alias="blam")
        not_init: int = attrs.field(init=False)
        no: bytes = attrs.field()

    old_model = StubModel(foo=42, baaaa="sheep", blam=True, no=b"okokokok")
    old_model.not_init = 54234

    copier = attrs_extensions.generate_shallow_copier(StubModel)
    new_model = copier(old_model)

    assert new_model is not old_model
    assert new_model._foo is old_model._foo
    assert new_model.baaaa is old_model.baaaa
    assert new_model._blam is old_model._blam
    assert new_model.not_init is old_model.not_init
    assert new_model.no is old_model.no


def test_generate_shallow_copier_with_init_only_arguments():
    @attrs.define()
    class StubModel:
        _gfd: int = attrs.field(alias="gfd")
        baaaa: str = attrs.field()
        _blambat: bool = attrs.field(alias="blambat")
        no: bytes = attrs.field()

    old_model = StubModel(gfd=42, baaaa="sheep", blambat=True, no=b"okokokok")

    copier = attrs_extensions.generate_shallow_copier(StubModel)
    new_model = copier(old_model)

    assert new_model is not old_model
    assert new_model._gfd is old_model._gfd
    assert new_model.baaaa is old_model.baaaa
    assert new_model._blambat is old_model._blambat
    assert new_model.no is old_model.no


def test_generate_shallow_copier_with_only_non_init_attrs():
    @attrs.define()
    class StubModel:
        _gfd: int = attrs.field(init=False)
        baaaa: str = attrs.field(init=False)
        _blambat: bool = attrs.field(init=False)
        no: bytes = attrs.field(init=False)

    old_model = StubModel()
    old_model._gfd = 42
    old_model.baaaa = "sheep"
    old_model._blambat = True
    old_model.no = b"okokokok"

    copier = attrs_extensions.generate_shallow_copier(StubModel)
    new_model = copier(old_model)

    assert new_model is not old_model
    assert new_model._gfd is old_model._gfd
    assert new_model.baaaa is old_model.baaaa
    assert new_model._blambat is old_model._blambat
    assert new_model.no is old_model.no


def test_generate_shallow_copier_with_no_attributes():
    @attrs.define()
    class StubModel: ...

    old_model = StubModel()

    copier = attrs_extensions.generate_shallow_copier(StubModel)
    new_model = copier(old_model)

    assert new_model is not old_model
    assert isinstance(new_model, StubModel)


def test_get_or_generate_shallow_copier_for_cached_copier():
    mock_copier = object()

    @attrs.define()
    class StubModel: ...

    attrs_extensions._SHALLOW_COPIERS = {
        type("b", (), {}): object(),
        StubModel: mock_copier,
        type("a", (), {}): object(),
    }

    assert attrs_extensions.get_or_generate_shallow_copier(StubModel) is mock_copier


def test_get_or_generate_shallow_copier_for_uncached_copier():
    mock_copier = object()

    @attrs.define()
    class StubModel: ...

    with mock.patch.object(attrs_extensions, "generate_shallow_copier", return_value=mock_copier):
        assert attrs_extensions.get_or_generate_shallow_copier(StubModel) is mock_copier

        attrs_extensions.generate_shallow_copier.assert_called_once_with(StubModel)

    assert attrs_extensions._SHALLOW_COPIERS[StubModel] is mock_copier


def test_copy_attrs():
    mock_result = object()
    mock_copier = mock.Mock(return_value=mock_result)

    @attrs.define()
    class StubModel: ...

    model = StubModel()

    with mock.patch.object(attrs_extensions, "get_or_generate_shallow_copier", return_value=mock_copier):
        assert attrs_extensions.copy_attrs(model) is mock_result

        attrs_extensions.get_or_generate_shallow_copier.assert_called_once_with(StubModel)

    mock_copier.assert_called_once_with(model)


def test_generate_deep_copier():
    @attrs.define
    class StubBaseClass:
        recursor: int = attrs.field()
        _field: bool = attrs.field(alias="field")
        foo: str = attrs.field()
        end: str = attrs.field(init=False)
        _blam: bool = attrs.field(init=False)

    model = StubBaseClass(recursor=431, field=True, foo="blam")
    model.end = "the way"
    model._blam = "555555"
    old_model_fields = stdlib_copy.copy(model)
    copied_recursor = object()
    copied_field = object()
    copied_foo = object()
    copied_end = object()
    copied_blam = object()
    memo = {123: object()}

    with mock.patch.object(
        stdlib_copy, "deepcopy", side_effect=[copied_recursor, copied_field, copied_foo, copied_end, copied_blam]
    ):
        attrs_extensions.generate_deep_copier(StubBaseClass)(model, memo)

        stdlib_copy.deepcopy.assert_has_calls(
            [
                mock.call(old_model_fields.recursor, memo),
                mock.call(old_model_fields._field, memo),
                mock.call(old_model_fields.foo, memo),
                mock.call(old_model_fields.end, memo),
                mock.call(old_model_fields._blam, memo),
            ]
        )

    assert model.recursor is copied_recursor
    assert model._field is copied_field
    assert model.foo is copied_foo
    assert model.end is copied_end
    assert model._blam is copied_blam


def test_generate_deep_copier_with_only_init_attributes():
    @attrs.define
    class StubBaseClass:
        recursor: int = attrs.field()
        _field: bool = attrs.field(alias="field")
        foo: str = attrs.field()

    model = StubBaseClass(recursor=431, field=True, foo="blam")
    old_model_fields = stdlib_copy.copy(model)
    copied_recursor = object()
    copied_field = object()
    copied_foo = object()
    memo = {123: object()}

    with mock.patch.object(stdlib_copy, "deepcopy", side_effect=[copied_recursor, copied_field, copied_foo]):
        attrs_extensions.generate_deep_copier(StubBaseClass)(model, memo)

        stdlib_copy.deepcopy.assert_has_calls(
            [
                mock.call(old_model_fields.recursor, memo),
                mock.call(old_model_fields._field, memo),
                mock.call(old_model_fields.foo, memo),
            ]
        )

    assert model.recursor is copied_recursor
    assert model._field is copied_field
    assert model.foo is copied_foo


def test_generate_deep_copier_with_only_non_init_attributes():
    @attrs.define
    class StubBaseClass:
        end: str = attrs.field(init=False)
        _blam: bool = attrs.field(init=False)

    model = StubBaseClass()
    model.end = "the way"
    model._blam = "555555"
    old_model_fields = stdlib_copy.copy(model)
    copied_end = object()
    copied_blam = object()
    memo = {123: object()}

    with mock.patch.object(stdlib_copy, "deepcopy", side_effect=[copied_end, copied_blam]):
        attrs_extensions.generate_deep_copier(StubBaseClass)(model, memo)

        stdlib_copy.deepcopy.assert_has_calls(
            [mock.call(old_model_fields.end, memo), mock.call(old_model_fields._blam, memo)]
        )

    assert model.end is copied_end
    assert model._blam is copied_blam


def test_generate_deep_copier_with_no_attributes():
    @attrs.define
    class StubBaseClass: ...

    model = StubBaseClass()
    memo = {123: object()}

    with mock.patch.object(stdlib_copy, "deepcopy", side_effect=NotImplementedError):
        attrs_extensions.generate_deep_copier(StubBaseClass)(model, memo)

        stdlib_copy.deepcopy.assert_not_called()


def test_get_or_generate_deep_copier_for_cached_function():
    class StubClass: ...

    mock_copier = object()
    attrs_extensions._DEEP_COPIERS = {}

    with mock.patch.object(attrs_extensions, "generate_deep_copier", return_value=mock_copier):
        assert attrs_extensions.get_or_generate_deep_copier(StubClass) is mock_copier

        attrs_extensions.generate_deep_copier.assert_called_once_with(StubClass)

    assert attrs_extensions._DEEP_COPIERS[StubClass] is mock_copier


def test_get_or_generate_deep_copier_for_uncached_function():
    class StubClass: ...

    mock_copier = object()
    attrs_extensions._DEEP_COPIERS = {StubClass: mock_copier}

    with mock.patch.object(attrs_extensions, "generate_deep_copier"):
        assert attrs_extensions.get_or_generate_deep_copier(StubClass) is mock_copier

        attrs_extensions.generate_deep_copier.assert_not_called()


def test_deep_copy_attrs_without_memo():
    class StubClass: ...

    mock_object = StubClass()
    mock_result = object()
    mock_copier = mock.Mock(mock_result)

    stack = contextlib.ExitStack()
    stack.enter_context(mock.patch.object(attrs_extensions, "get_or_generate_deep_copier", return_value=mock_copier))
    stack.enter_context(mock.patch.object(stdlib_copy, "copy", return_value=mock_result))

    with stack:
        assert attrs_extensions.deep_copy_attrs(mock_object) is mock_result

        stdlib_copy.copy.assert_called_once_with(mock_object)
        attrs_extensions.get_or_generate_deep_copier.assert_called_once_with(StubClass)

    mock_copier.assert_called_once_with(mock_result, {id(mock_object): mock_result})


def test_deep_copy_attrs_with_memo():
    class StubClass: ...

    mock_object = StubClass()
    mock_result = object()
    mock_copier = mock.Mock(mock_result)
    mock_other_object = object()

    stack = contextlib.ExitStack()
    stack.enter_context(mock.patch.object(attrs_extensions, "get_or_generate_deep_copier", return_value=mock_copier))
    stack.enter_context(mock.patch.object(stdlib_copy, "copy", return_value=mock_result))

    with stack:
        assert attrs_extensions.deep_copy_attrs(mock_object, {1235342: mock_other_object}) is mock_result

        stdlib_copy.copy.assert_called_once_with(mock_object)
        attrs_extensions.get_or_generate_deep_copier.assert_called_once_with(StubClass)

    mock_copier.assert_called_once_with(mock_result, {id(mock_object): mock_result, 1235342: mock_other_object})


class TestCopyDecorator:
    def test___copy__(self):
        mock_result = object()
        mock_copier = mock.Mock(return_value=mock_result)

        @attrs.define()
        @attrs_extensions.with_copy
        class StubClass: ...

        model = StubClass()

        with mock.patch.object(attrs_extensions, "get_or_generate_shallow_copier", return_value=mock_copier):
            assert stdlib_copy.copy(model) is mock_result

            attrs_extensions.get_or_generate_shallow_copier.assert_called_once_with(StubClass)

        mock_copier.assert_called_once_with(model)

    def test___deep__copy(self):
        class CopyingMock(mock.Mock):
            def __call__(self, /, *args, **kwargs):
                args = list(args)
                args[1] = dict(args[1])
                return super().__call__(*args, **kwargs)

        mock_result = object()
        mock_copier = CopyingMock(return_value=mock_result)

        @attrs.define()
        @attrs_extensions.with_copy
        class StubClass: ...

        model = StubClass()
        stack = contextlib.ExitStack()
        stack.enter_context(
            mock.patch.object(attrs_extensions, "get_or_generate_deep_copier", return_value=mock_copier)
        )
        stack.enter_context(mock.patch.object(stdlib_copy, "copy", return_value=mock_result))

        with stack:
            assert stdlib_copy.deepcopy(model) is mock_result

            stdlib_copy.copy.assert_called_once_with(model)
            attrs_extensions.get_or_generate_deep_copier.assert_called_once_with(StubClass)

        mock_copier.assert_called_once_with(mock_result, {id(model): mock_result})

    def test_copy_decorator_inheritance(self):
        @attrs_extensions.with_copy
        @attrs.define()
        class ParentClass: ...

        class Foo(ParentClass): ...

        assert Foo.__copy__ == attrs_extensions.copy_attrs
        assert Foo.__deepcopy__ == attrs_extensions.deep_copy_attrs



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Base types and functions for events in Hikari."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "Event",
    "EventT",
    "ExceptionEvent",
    "get_required_intents_for",
    "is_no_recursive_throw_event",
    "no_recursive_throw",
    "requires_intents",
)

import abc
import inspect
import typing

import attrs

from hikari.api import shard as gateway_shard
from hikari.internal import attrs_extensions
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    import types

    from hikari import intents
    from hikari import traits

    _T = typing.TypeVar("_T")

REQUIRED_INTENTS_attrs: typing.Final[str] = "___requiresintents___"
NO_RECURSIVE_THROW_attrs: typing.Final[str] = "___norecursivethrow___"

_id_counter = 1  # We start at 1 since Event is 0


class Event(abc.ABC):
    """Base event type that all Hikari events should subclass."""

    __slots__: typing.Sequence[str] = ()

    __dispatches: typing.ClassVar[tuple[type[Event], ...]]
    __bitmask: typing.ClassVar[int]

    @typing_extensions.override
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if not hasattr(Event, "__dispatches"):
            Event.__dispatches = (Event,)
            Event.__bitmask = 1 << 0

        global _id_counter  # noqa: PLW0603 - Do use use global

        mro = cls.mro()
        # We don't have to explicitly include Event here as issubclass(Event, Event) returns True.
        # Non-event classes should be ignored.
        cls.__dispatches = tuple(sub_cls for sub_cls in mro if issubclass(sub_cls, Event))
        cls.__bitmask = 1 << _id_counter
        _id_counter += 1

    @property
    @abc.abstractmethod
    def app(self) -> traits.RESTAware:
        """App instance for this application."""

    @classmethod
    def dispatches(cls) -> typing.Sequence[type[Event]]:
        """Sequence of the event classes this event is dispatched as."""
        return cls.__dispatches

    @classmethod
    def bitmask(cls) -> int:
        """Bitmask for this event."""
        return cls.__bitmask


def get_required_intents_for(event_type: type[Event]) -> typing.Collection[intents.Intents]:
    """Retrieve the intents that are required to listen to an event type.

    Parameters
    ----------
    event_type
        The event type to get required intents for.

    Returns
    -------
    typing.Collection[hikari.intents.Intents]
        Collection of acceptable subset combinations of intent needed to
        be able to receive the given event type.
    """
    result = getattr(event_type, REQUIRED_INTENTS_attrs, ())
    assert isinstance(result, typing.Collection)
    return result


def requires_intents(first: intents.Intents, *rest: intents.Intents) -> typing.Callable[[_T], _T]:
    """Decorate an event type to define what intents it requires.

    Parameters
    ----------
    first
        First combination of intents that are acceptable in order to receive
        the decorated event type.
    *rest
        Zero or more additional combinations of intents to require for this
        event to be subscribed to.
    """

    def decorator(cls: _T) -> _T:
        required_intents = [first, *rest]
        setattr(cls, REQUIRED_INTENTS_attrs, required_intents)
        doc = inspect.getdoc(cls) or ""
        doc += "\n\nThis requires one of the following combinations of intents in order to be dispatched:\n\n"
        for intent_group in required_intents:
            preview = " + ".join(
                f"`{type(i).__module__}.{type(i).__qualname__}.{i.name}`" for i in intent_group.split()
            )
            doc += f" - {preview}\n"

        cls.__doc__ = doc
        return cls

    return decorator


def no_recursive_throw() -> typing.Callable[[_T], _T]:
    """Decorate an event type to indicate errors should not be handled.

    This is useful for exception event types that you do not want to
    have invoked recursively.
    """

    def decorator(cls: _T) -> _T:
        setattr(cls, NO_RECURSIVE_THROW_attrs, True)
        doc = inspect.getdoc(cls) or ""
        doc += (
            "\n"
            "!!! warning\n"
            "    Any exceptions raised by handlers for this event will be dumped to the\n"
            "    application logger and silently discarded, preventing recursive loops\n"
            "    produced by faulty exception event handling. Thus, it is imperative\n"
            "    that you ensure any exceptions are explicitly caught within handlers\n"
            "    for this event if they may occur.\n"
        )
        cls.__doc__ = doc
        return cls

    return decorator


def is_no_recursive_throw_event(obj: _T | type[_T]) -> bool:
    """Whether the event is marked as `___norecursivethrow___`."""
    result = getattr(obj, NO_RECURSIVE_THROW_attrs, False)
    assert isinstance(result, bool)
    return result


EventT = typing.TypeVar("EventT", bound=Event)
FailedCallbackT = typing.Callable[[EventT], typing.Coroutine[typing.Any, typing.Any, None]]


@no_recursive_throw()
@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ExceptionEvent(Event, typing.Generic[EventT]):
    """Event that is raised when another event handler raises an [`Exception`][].

    !!! note
        Only exceptions that derive from [`Exception`][] will be caught.
        Other exceptions outside this range will propagate past this callback.
        This prevents event handlers interfering with critical exceptions
        such as [`KeyboardInterrupt`][] which would have potentially undesired
        side-effects on the application runtime.
    """

    exception: Exception = attrs.field()
    """Exception that was raised."""

    failed_event: EventT = attrs.field()
    """Event instance that caused the exception."""

    failed_callback: FailedCallbackT[EventT] = attrs.field()
    """Event callback that threw an exception."""

    @property
    @typing_extensions.override
    def app(self) -> traits.RESTAware:
        # <<inherited docstring from Event>>.
        return self.failed_event.app

    @property
    def shard(self) -> gateway_shard.GatewayShard | None:
        """Shard that received the event, if there was one associated.

        This may be [`None`][] if no specific shard was the cause of this
        exception (e.g. when starting up or shutting down).
        """
        shard = getattr(self.failed_event, "shard", None)
        if isinstance(shard, gateway_shard.GatewayShard):
            return shard
        return None

    @property
    def exc_info(self) -> tuple[type[Exception], Exception, types.TracebackType | None]:
        """Exception triplet that follows the same format as [`sys.exc_info`][].

        The [`sys.exc_info`][] triplet consists of the exception type, the exception
        instance, and the traceback of the exception.
        """
        return type(self.exception), self.exception, self.exception.__traceback__

    async def retry(self) -> None:
        """Invoke the failed event again.

        If an exception is thrown this time, it will need to be manually
        caught in-code, or will be discarded.
        """
        await self.failed_callback(self.failed_event)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import snowflakes
from hikari import traits
from hikari import undefined
from hikari import urls
from hikari import users
from hikari.internal import routes
from tests.hikari import hikari_test_helpers


class TestPartialUser:
    @pytest.fixture
    def obj(self):
        # ABC, so must be stubbed.
        return hikari_test_helpers.mock_class_namespace(users.PartialUser, slots_=False)()

    def test_accent_colour_alias_property(self, obj):
        obj.accent_color = object()

        assert obj.accent_colour is obj.accent_color

    @pytest.mark.asyncio
    async def test_fetch_self(self, obj):
        obj.id = 123
        obj.app = mock.AsyncMock()

        assert await obj.fetch_self() is obj.app.rest.fetch_user.return_value
        obj.app.rest.fetch_user.assert_awaited_once_with(user=123)

    @pytest.mark.asyncio
    async def test_send_uses_cached_id(self, obj):
        obj.id = 4123123
        embed = object()
        embeds = [object()]
        attachment = object()
        attachments = [object(), object()]
        component = object()
        components = [object(), object()]
        user_mentions = [object(), object()]
        role_mentions = [object(), object()]
        reply = object()
        mentions_reply = object()

        obj.app = mock.Mock(spec=traits.CacheAware, rest=mock.AsyncMock())
        obj.fetch_dm_channel = mock.AsyncMock()

        returned = await obj.send(
            content="test",
            embed=embed,
            embeds=embeds,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            tts=True,
            reply=reply,
            reply_must_exist=False,
            mentions_everyone=False,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
            mentions_reply=mentions_reply,
            flags=34123342,
        )

        assert returned is obj.app.rest.create_message.return_value

        obj.app.cache.get_dm_channel_id.assert_called_once_with(4123123)
        obj.fetch_dm_channel.assert_not_called()
        obj.app.rest.create_message.assert_awaited_once_with(
            channel=obj.app.cache.get_dm_channel_id.return_value,
            content="test",
            embed=embed,
            embeds=embeds,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            tts=True,
            mentions_everyone=False,
            reply_must_exist=False,
            reply=reply,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
            mentions_reply=mentions_reply,
            flags=34123342,
        )

    @pytest.mark.asyncio
    async def test_send_when_not_cached(self, obj):
        obj.id = 522234
        obj.app = mock.Mock(spec=traits.CacheAware, rest=mock.AsyncMock())
        obj.app.cache.get_dm_channel_id = mock.Mock(return_value=None)
        obj.fetch_dm_channel = mock.AsyncMock()

        returned = await obj.send()

        assert returned is obj.app.rest.create_message.return_value

        obj.app.cache.get_dm_channel_id.assert_called_once_with(522234)
        obj.fetch_dm_channel.assert_awaited_once()
        obj.app.rest.create_message.assert_awaited_once_with(
            channel=obj.fetch_dm_channel.return_value.id,
            content=undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            tts=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            reply=undefined.UNDEFINED,
            reply_must_exist=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
            mentions_reply=undefined.UNDEFINED,
            flags=undefined.UNDEFINED,
        )

    @pytest.mark.asyncio
    async def test_send_when_not_cache_aware(self, obj):
        obj.id = 522234
        obj.app = mock.Mock(spec=traits.RESTAware, rest=mock.AsyncMock())
        obj.fetch_dm_channel = mock.AsyncMock()

        returned = await obj.send()

        assert returned is obj.app.rest.create_message.return_value

        obj.fetch_dm_channel.assert_awaited_once()
        obj.app.rest.create_message.assert_awaited_once_with(
            channel=obj.fetch_dm_channel.return_value.id,
            content=undefined.UNDEFINED,
            embed=undefined.UNDEFINED,
            embeds=undefined.UNDEFINED,
            attachment=undefined.UNDEFINED,
            attachments=undefined.UNDEFINED,
            component=undefined.UNDEFINED,
            components=undefined.UNDEFINED,
            tts=undefined.UNDEFINED,
            mentions_everyone=undefined.UNDEFINED,
            reply=undefined.UNDEFINED,
            reply_must_exist=undefined.UNDEFINED,
            user_mentions=undefined.UNDEFINED,
            role_mentions=undefined.UNDEFINED,
            mentions_reply=undefined.UNDEFINED,
            flags=undefined.UNDEFINED,
        )

    @pytest.mark.asyncio
    async def test_fetch_dm_channel(self, obj):
        obj.id = 123
        obj.app = mock.Mock()
        obj.app.rest.create_dm_channel = mock.AsyncMock()

        assert await obj.fetch_dm_channel() == obj.app.rest.create_dm_channel.return_value

        obj.app.rest.create_dm_channel.assert_awaited_once_with(123)


class TestUser:
    @pytest.fixture
    def obj(self):
        # ABC, so must be stubbed.
        return hikari_test_helpers.mock_class_namespace(users.User, slots_=False)()

    def test_accent_colour_alias_property(self, obj):
        obj.accent_color = object()

        assert obj.accent_colour is obj.accent_color

    def test_avatar_decoration_property(self, obj):
        obj.avatar_decoration = users.AvatarDecoration(
            asset_hash="18dnf8dfbakfdh", sku_id=snowflakes.Snowflake(123), expires_at=None
        )

        with mock.patch.object(users.AvatarDecoration, "make_url") as make_url:
            assert obj.avatar_decoration.url is make_url.return_value

    def test_avatar_decoration_make_url_with_all_args(self, obj):
        obj.avatar_decoration = users.AvatarDecoration(
            asset_hash="18dnf8dfbakfdh", sku_id=snowflakes.Snowflake(123), expires_at=None
        )

        with mock.patch.object(
            routes, "CDN_AVATAR_DECORATION", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.avatar_decoration.make_url(size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.MEDIA_PROXY_URL, hash=obj.avatar_decoration.asset_hash, size=4096, file_format="PNG", lossless=True
        )

    def test_make_avatar_url_format_set_to_deprecated_ext_argument_if_provided(self, obj):
        obj.id = 123321
        obj.avatar_hash = "fofoof"

        with mock.patch.object(
            routes, "CDN_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_avatar_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=123321, hash="fofoof", size=4096, file_format="JPEG", lossless=True
        )

    def test_avatar_url_property(self, obj):
        with mock.patch.object(users.User, "make_avatar_url") as make_avatar_url:
            assert obj.avatar_url is make_avatar_url.return_value

    def test_make_avatar_url_when_no_hash(self, obj):
        obj.avatar_hash = None
        assert obj.make_avatar_url(file_format="PNG", size=1024) is None

    def test_make_avatar_url_when_format_is_None_and_avatar_hash_is_for_gif(self, obj):
        obj.avatar_hash = "a_18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_avatar_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=obj.id, hash="a_18dnf8dfbakfdh", size=4096, file_format="GIF", lossless=True
        )

    def test_make_avatar_url_when_format_is_None_and_avatar_hash_is_not_for_gif(self, obj):
        obj.avatar_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_avatar_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=obj.id, hash=obj.avatar_hash, size=4096, file_format="PNG", lossless=True
        )

    def test_make_avatar_url_with_all_args(self, obj):
        obj.avatar_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_avatar_url(file_format="URL", size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=obj.id, hash=obj.avatar_hash, size=4096, file_format="URL", lossless=True
        )

    def test_display_avatar_url_when_avatar_url(self, obj):
        with mock.patch.object(users.User, "make_avatar_url") as mock_make_avatar_url:
            assert obj.display_avatar_url is mock_make_avatar_url.return_value

    def test_display_avatar_url_when_no_avatar_url(self, obj):
        with mock.patch.object(users.User, "make_avatar_url", return_value=None):
            with mock.patch.object(users.User, "default_avatar_url") as mock_default_avatar_url:
                assert obj.display_avatar_url is mock_default_avatar_url

    def test_display_banner_url_when_banner_url(self, obj):
        with mock.patch.object(users.User, "make_banner_url") as mock_make_banner_url:
            assert obj.display_banner_url is mock_make_banner_url.return_value

    def test_display_banner_url_when_no_banner_url(self, obj):
        with mock.patch.object(users.User, "make_banner_url", return_value=None):
            assert obj.display_banner_url is None

    def test_default_avatar(self, obj):
        obj.avatar_hash = "18dnf8dfbakfdh"
        obj.discriminator = "1234"

        with mock.patch.object(
            routes, "CDN_DEFAULT_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.default_avatar_url == "file"

        route.compile_to_file.assert_called_once_with(urls.CDN_URL, style=4, file_format="PNG")

    def test_default_avatar_for_migrated_users(self, obj):
        obj.id = 377812572784820226
        obj.avatar_hash = "18dnf8dfbakfdh"
        obj.discriminator = "0"

        with mock.patch.object(
            routes, "CDN_DEFAULT_USER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.default_avatar_url == "file"

        route.compile_to_file.assert_called_once_with(urls.CDN_URL, style=0, file_format="PNG")

    def test_make_banner_url_format_set_to_deprecated_ext_argument_if_provided(self, obj):
        obj.id = 123321
        obj.banner_hash = "fofoof"

        with mock.patch.object(
            routes, "CDN_USER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_banner_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=123321, hash="fofoof", size=4096, file_format="JPEG", lossless=True
        )

    def test_banner_url_property(self, obj):
        with mock.patch.object(users.User, "make_banner_url") as make_banner_url:
            assert obj.banner_url is make_banner_url.return_value

    def test_make_banner_url_when_no_hash(self, obj):
        obj.banner_hash = None

        with mock.patch.object(routes, "CDN_USER_BANNER") as route:
            assert obj.make_banner_url(file_format=None, size=4096) is None

        route.compile_to_file.assert_not_called()

    def test_make_banner_url_when_format_is_None_and_banner_hash_is_for_gif(self, obj):
        obj.banner_hash = "a_18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_USER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_banner_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=obj.id, hash="a_18dnf8dfbakfdh", size=4096, file_format="GIF", lossless=True
        )

    def test_make_banner_url_when_format_is_None_and_banner_hash_is_not_for_gif(self, obj):
        obj.banner_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_USER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_banner_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=obj.id, hash=obj.banner_hash, size=4096, file_format="PNG", lossless=True
        )

    def test_make_banner_url_with_all_args(self, obj):
        obj.banner_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_USER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert obj.make_banner_url(file_format="URL", size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, user_id=obj.id, hash=obj.banner_hash, size=4096, file_format="URL", lossless=True
        )


class TestPartialUserImpl:
    @pytest.fixture
    def obj(self):
        return users.PartialUserImpl(
            id=snowflakes.Snowflake(123),
            app=mock.Mock(),
            discriminator="8637",
            username="thomm.o",
            global_name=None,
            avatar_decoration=None,
            avatar_hash=None,
            banner_hash=None,
            accent_color=None,
            is_bot=False,
            is_system=False,
            flags=users.UserFlag.DISCORD_EMPLOYEE,
        )

    def test_str_operator(self, obj):
        assert str(obj) == "thomm.o#8637"

    def test_str_operator_when_partial(self, obj):
        obj.username = undefined.UNDEFINED
        assert str(obj) == "Partial user ID 123"

    def test_mention_property(self, obj):
        assert obj.mention == "<@123>"

    def test_display_name_property_when_global_name(self, obj):
        obj.global_name = "Thommo"
        assert obj.display_name == obj.global_name

    def test_display_name_property_when_no_global_name(self, obj):
        obj.global_name = None
        assert obj.display_name == obj.username

    @pytest.mark.asyncio
    async def test_fetch_self(self, obj):
        user = object()
        obj.app.rest.fetch_user = mock.AsyncMock(return_value=user)
        assert await obj.fetch_self() is user
        obj.app.rest.fetch_user.assert_awaited_once_with(user=123)


@pytest.mark.asyncio
class TestOwnUser:
    @pytest.fixture
    def obj(self):
        return users.OwnUser(
            id=snowflakes.Snowflake(12345),
            app=mock.Mock(),
            discriminator="1234",
            username="foobar",
            global_name=None,
            avatar_decoration=None,
            avatar_hash="69420",
            banner_hash="42069",
            accent_color=123456,
            is_bot=False,
            is_system=False,
            flags=users.UserFlag.PARTNERED_SERVER_OWNER,
            is_mfa_enabled=True,
            locale="en-GB",
            is_verified=False,
            email="someone@example.com",
            premium_type=None,
        )

    async def test_fetch_self(self, obj):
        user = object()
        obj.app.rest.fetch_my_user = mock.AsyncMock(return_value=user)
        assert await obj.fetch_self() is user
        obj.app.rest.fetch_my_user.assert_awaited_once_with()

    async def test_fetch_dm_channel(self, obj):
        with pytest.raises(TypeError, match=r"Unable to fetch your own DM channel"):
            await obj.fetch_dm_channel()

    async def test_send(self, obj):
        with pytest.raises(TypeError, match=r"Unable to send a DM to yourself"):
            await obj.send()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import copy

import mock

from hikari import snowflakes
from hikari import stickers
from hikari.internal import cache


class TestStickerData:
    def test_from_entity(self) -> None:
        mock_user = object()
        mock_sticker = stickers.GuildSticker(
            id=snowflakes.Snowflake(69420),
            name="lulzor",
            format_type=123,
            description="Fake sticker",
            guild_id=snowflakes.Snowflake(42069),
            is_available=True,
            tag="some tag",
            user=None,
        )

        data = cache.GuildStickerData.build_from_entity(mock_sticker, user=mock_user)

        assert data.id is mock_sticker.id
        assert data.name is mock_sticker.name
        assert data.format_type is mock_sticker.format_type
        assert data.description is mock_sticker.description
        assert data.guild_id is mock_sticker.guild_id
        assert data.is_available is mock_sticker.is_available
        assert data.tag is mock_sticker.tag
        assert data.user is mock_user

    def test_from_entity_when_user_not_passed(self) -> None:
        mock_user = object()
        mock_sticker = mock_sticker = stickers.GuildSticker(
            id=snowflakes.Snowflake(69420),
            name="lulzor",
            format_type=123,
            description="Fake sticker",
            guild_id=snowflakes.Snowflake(42069),
            is_available=True,
            tag="some tag",
            user=mock_user,
        )

        with mock.patch.object(copy, "copy") as mock_copy:
            with mock.patch.object(cache, "RefCell") as refcell:
                data = cache.GuildStickerData.build_from_entity(mock_sticker)

        assert data.user is refcell.return_value
        mock_copy.assert_called_once_with(mock_user)
        refcell.assert_called_once_with(mock_copy.return_value)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Data binding utilities."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "Headers",
    "JSONArray",
    "JSONObject",
    "JSONObjectBuilder",
    "JSONPayload",
    "JSONish",
    "Query",
    "StringMapBuilder",
    "URLEncodedFormBuilder",
    "default_json_dumps",
    "default_json_loads",
)

import datetime
import typing

import aiohttp
import multidict

from hikari import errors
from hikari import files
from hikari import snowflakes
from hikari import undefined

if typing.TYPE_CHECKING:
    import concurrent.futures
    import contextlib

    T_co = typing.TypeVar("T_co", covariant=True)
    T = typing.TypeVar("T")

Headers = typing.Mapping[str, str]
"""Type hint for HTTP headers."""

Query = typing.Union[dict[str, str], multidict.MultiDict[str]]
"""Type hint for HTTP query string."""

# MyPy does not support recursive types yet. This has been ongoing for a long time, unfortunately.
# See https://github.com/python/typing/issues/182

JSONObject = typing.Mapping[str, typing.Any]
"""Type hint for a JSON-decoded object representation as a mapping."""

JSONArray = typing.Sequence[typing.Any]
"""Type hint for a JSON-decoded array representation as a sequence."""

JSONish = typing.Union[str, int, float, bool, None, JSONArray, JSONObject]
"""Type hint for any valid JSON-decoded type."""

Stringish = typing.Union[str, int, bool, datetime.datetime, undefined.UndefinedType, None, snowflakes.Unique]
"""Type hint for any valid that can be put in a StringMapBuilder"""

JSONEncoder = typing.Callable[[typing.Union[JSONArray, JSONObject]], bytes]
"""Type hint for hikari-compatible JSON encoders.

A hikari-compatible JSON encoder is one which will take in a JSON-ish object and output either [`str`][]
or [`bytes`][].
"""

JSONDecoder = typing.Callable[[typing.Union[str, bytes]], typing.Union[JSONArray, JSONObject]]
"""Type hint for hikari-compatible JSON decoder.

A hikari-compatible JSON decoder is one which will take either a [`str`][] or [`bytes`][] and outputs
the JSON-ish object, as well as raises a [`ValueError`][] on an incorrect JSON payload being passed in.
"""

_StringMapBuilderArg = typing.Union[
    typing.Mapping[str, str], multidict.MultiMapping[str], typing.Iterable[tuple[str, str]]
]

_APPLICATION_OCTET_STREAM: typing.Final[str] = "application/octet-stream"
_JSON_CONTENT_TYPE: typing.Final[str] = "application/json"
_UTF_8: typing.Final[str] = "utf-8"

default_json_dumps: JSONEncoder
"""Default JSON encoder to use."""

default_json_loads: JSONDecoder
"""Default JSON decoder to use."""

try:
    import orjson

    def default_json_dumps(obj: JSONArray | JSONObject) -> bytes:
        """Encode a JSON object to [`bytes`][]."""
        return orjson.dumps(obj, option=orjson.OPT_NON_STR_KEYS)

    default_json_loads = orjson.loads
except ModuleNotFoundError:
    import json

    _json_separators = (",", ":")

    def default_json_dumps(obj: JSONArray | JSONObject) -> bytes:
        """Encode a JSON object to [`bytes`][]."""
        return json.dumps(obj, separators=_json_separators).encode(_UTF_8)

    default_json_loads = json.loads


@typing.final
class JSONPayload(aiohttp.BytesPayload):
    """A JSON payload to use in an aiohttp request."""

    def __init__(self, value: JSONArray | JSONObject, dumps: JSONEncoder = default_json_dumps) -> None:
        super().__init__(dumps(value), content_type=_JSON_CONTENT_TYPE, encoding=_UTF_8)


@typing.final
class URLEncodedFormBuilder:
    """Helper class to generate [`aiohttp.FormData`][]."""

    __slots__: typing.Sequence[str] = ("_fields", "_resources")

    def __init__(self) -> None:
        self._fields: list[tuple[str, str | aiohttp.BytesPayload, str | None]] = []
        self._resources: list[tuple[str, files.Resource[files.AsyncReader]]] = []

    def add_field(self, name: str, data: str | bytes, *, content_type: str | None = None) -> None:
        field_data: str | aiohttp.BytesPayload = aiohttp.BytesPayload(data) if isinstance(data, bytes) else data
        self._fields.append((name, field_data, content_type))

    def add_resource(self, name: str, resource: files.Resource[files.AsyncReader]) -> None:
        self._resources.append((name, resource))

    async def build(
        self, stack: contextlib.AsyncExitStack, executor: concurrent.futures.Executor | None = None
    ) -> aiohttp.FormData:
        form = aiohttp.FormData()

        for field in self._fields:
            form.add_field(field[0], field[1], content_type=field[2])

        for name, resource in self._resources:
            stream = await stack.enter_async_context(resource.stream(executor=executor))
            mimetype = stream.mimetype or _APPLICATION_OCTET_STREAM
            form.add_field(name, stream, filename=stream.filename, content_type=mimetype)

        return form


@typing.final
class StringMapBuilder(multidict.MultiDict[str]):
    """Helper class used to quickly build query strings or header maps.

    This will consume any items that are not [`hikari.undefined.UNDEFINED`][].
    If a value is unspecified, it will be ignored when inserting it. This reduces
    the amount of boilerplate needed for generating the headers and query strings for
    low-level HTTP API interaction, amongst other things.

    !!! warning
        Because this subclasses [`dict`][], you should not use the
        index operator to set items on this object. Doing so will skip any
        form of validation on the type. Use the [put*][] methods instead.
    """

    __slots__: typing.Sequence[str] = ()

    def __init__(self, arg: _StringMapBuilderArg = (), **kwargs: str) -> None:
        # We have to allow arguments to be passed to the init here otherwise the inherited copy behaviour from
        # multidict.MultiDict fails.
        super().__init__(arg, **kwargs)

    @typing.overload
    def put(self, key: str, value: Stringish, /) -> None: ...

    @typing.overload
    def put(
        self, key: str, value: undefined.UndefinedOr[T_co], /, *, conversion: typing.Callable[[T_co], Stringish]
    ) -> None: ...

    def put(
        self,
        key: str,
        value: undefined.UndefinedOr[typing.Any],
        /,
        *,
        conversion: typing.Callable[[typing.Any], typing.Any] | None = None,
    ) -> None:
        """Add a key and value to the string map.

        !!! note
            The value will always be cast to a [`str`][] before inserting it.
            [`True`][] will be translated to `"true"`, [`False`][] will be
            translated to `"false"`, and [`None`][] will be translated to
            `"null"`.

        Parameters
        ----------
        key
            The string key.
        value
            The value to set.
        conversion
            An optional conversion to perform.
        """
        if value is undefined.UNDEFINED:
            return

        if conversion is not None:
            value = conversion(value)

        if value is True:
            value = "true"
        elif value is False:
            value = "false"
        elif value is None:
            value = "null"
        elif isinstance(value, snowflakes.Unique):
            value = str(value.id)
        elif isinstance(value, datetime.datetime):
            value = str(snowflakes.Snowflake.from_datetime(value))
        else:
            value = str(value)

        # __setitem__ just overwrites the previous value.
        self.add(key, value)


@typing.final
class JSONObjectBuilder(dict[str, JSONish]):
    """Helper class used to quickly build JSON objects from various values.

    If provided with any values that are [`hikari.undefined.UNDEFINED`][],
    then these values will be ignored.

    This speeds up generation of JSON payloads for low level HTTP and websocket
    API interaction.

    !!! warning
        Because this subclasses [`dict`][], you should not use the
        index operator to set items on this object. Doing so will skip any
        form of validation on the type. Use the [put*][] methods instead.
    """

    __slots__: typing.Sequence[str] = ()

    def __init__(self) -> None:
        # Only allow use of empty constructor here.
        super().__init__()

    @typing.overload
    def put(self, key: str, value: undefined.UndefinedNoneOr[JSONish], /) -> None: ...

    @typing.overload
    def put(
        self, key: str, value: undefined.UndefinedNoneOr[T_co], /, *, conversion: typing.Callable[[T_co], JSONish]
    ) -> None: ...

    def put(
        self,
        key: str,
        value: undefined.UndefinedNoneOr[typing.Any],
        /,
        *,
        conversion: typing.Callable[[typing.Any], JSONish] | None = None,
    ) -> None:
        """Put a JSON value.

        If the value is [`hikari.undefined.UNDEFINED`][] it will not be stored.

        Parameters
        ----------
        key
            The key to give the element.
        value
            The JSON type to put. This may be a non-JSON type if a conversion
            is also specified. This may alternatively be undefined. In the latter
            case, nothing is performed.
        conversion
            The optional conversion to apply.
        """
        if value is undefined.UNDEFINED:
            return

        if conversion is None or value is None:
            self[key] = value
        else:
            self[key] = conversion(value)

    @typing.overload
    def put_array(self, key: str, values: undefined.UndefinedOr[typing.Iterable[JSONish]], /) -> None: ...

    @typing.overload
    def put_array(
        self,
        key: str,
        values: undefined.UndefinedOr[typing.Iterable[T_co]],
        /,
        *,
        conversion: typing.Callable[[T_co], JSONish],
    ) -> None: ...

    def put_array(
        self,
        key: str,
        values: undefined.UndefinedOr[typing.Iterable[typing.Any]],
        /,
        *,
        conversion: typing.Callable[[typing.Any], JSONish] | None = None,
    ) -> None:
        """Put a JSON array.

        If the value is [`hikari.undefined.UNDEFINED`][] it will not be stored.

        If provided, a conversion will be applied to each item.

        Parameters
        ----------
        key
            The key to give the element.
        values
            The JSON types to put. This may be an iterable of non-JSON types if
            a conversion is also specified. This may alternatively be undefined.
            In the latter case, nothing is performed.
        conversion
            The optional conversion to apply.
        """
        if values is not undefined.UNDEFINED:
            if conversion is not None:
                self[key] = [conversion(value) for value in values]
            else:
                self[key] = list(values)

    def put_snowflake(
        self, key: str, value: undefined.UndefinedNoneOr[snowflakes.SnowflakeishOr[snowflakes.Unique]], /
    ) -> None:
        """Put a key with a snowflake value into the builder.

        If the value is [`hikari.undefined.UNDEFINED`][] it will not be stored.

        Parameters
        ----------
        key
            The key to give the element.
        value
            The JSON type to put. This may alternatively be undefined, in this
            case, nothing is performed. This may also be [`None`][], in this
            case the value isn't cast.
        """
        if value is not undefined.UNDEFINED and value is not None:
            self[key] = str(int(value))
        elif value is None:
            self[key] = None

    def put_snowflake_array(
        self, key: str, values: undefined.UndefinedOr[typing.Iterable[snowflakes.SnowflakeishOr[snowflakes.Unique]]], /
    ) -> None:
        """Put an array of snowflakes with the given key into this builder.

        If the value is [`hikari.undefined.UNDEFINED`][] it will not be stored.

        Each snowflake should be castable to an [`int`][].

        Parameters
        ----------
        key
            The key to give the element.
        values
            The JSON snowflakes to put. This may alternatively be undefined.
            In the latter case, nothing is performed.
        """
        if values is not undefined.UNDEFINED:
            self[key] = [str(int(value)) for value in values]


def cast_variants_array(cast: typing.Callable[[T_co], T], raw_values: typing.Iterable[T_co], /) -> list[T]:
    """Cast an array of enum variants while ignoring unrecognised variant types.

    Parameters
    ----------
    cast
        Callback to cast each variant to.

        This will ignore any variants which raises
        [`hikari.errors.UnrecognisedEntityError`][] on cast.
    raw_values
        Iterable of the raw values to cast.

    Returns
    -------
    typing.List[T]
        A list of the casted variants (with any unrecognised types ignored).
    """
    results: list[T] = []

    for value in raw_values:
        try:
            results.append(cast(value))

        except errors.UnrecognisedEntityError:  # noqa: PERF203 - Move try-except outside of loop
            pass

    return results



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import copy
import pickle

import pytest

from hikari import undefined


class TestUndefined:
    def test_repr(self):
        assert repr(undefined.UNDEFINED) == "UNDEFINED"

    def test_str(self):
        assert str(undefined.UNDEFINED) == "UNDEFINED"

    def test_bool(self):
        assert bool(undefined.UNDEFINED) is False

    def test_singleton_behaviour(self):
        assert undefined.UNDEFINED is undefined.UNDEFINED
        assert undefined.UNDEFINED == undefined.UNDEFINED
        assert undefined.UNDEFINED is not None
        assert undefined.UNDEFINED is not False

    def test_count(self):
        assert undefined.count(9, 18, undefined.UNDEFINED, 36, undefined.UNDEFINED, 54) == 2

    def test_cannot_reinstatiate(self):
        with pytest.raises(TypeError):
            type(undefined.UNDEFINED)()

    def test_copy(self):
        assert copy.copy(undefined.UNDEFINED) is undefined.UNDEFINED

    def test_deepcopy(self):
        assert copy.deepcopy(undefined.UNDEFINED) is undefined.UNDEFINED

    def test_can_pickle(self):
        data = pickle.dumps(undefined.UNDEFINED)
        deser = pickle.loads(data)  # noqa: S301 pickle loads is unsafe with untrusted data
        assert deser is undefined.UNDEFINED

    def test__getstate__(self):
        assert undefined.UNDEFINED.__getstate__() is False


@pytest.mark.parametrize(
    ("values", "result"),
    [
        ((), True),
        (("321", undefined.UNDEFINED, 33123), False),
        ((undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED), True),
        (("123", 23123, 34123, 312321, 543453, 432, 41, 231, 1243, 321), False),
        ((undefined.UNDEFINED,), True),
        ((undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, 34123), False),
    ],
)
def test_all_undefined(values, result):
    assert undefined.all_undefined(*values) is result


@pytest.mark.parametrize(
    ("values", "result"),
    [
        ((undefined.UNDEFINED,), True),
        ((), False),
        ((undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED), True),
        (("123", 432, 123, "43", 1223, "541324", 123453, 123, "543", 123), False),
        ((undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, "34r123"), True),
        ((undefined.UNDEFINED, 34123, 5432123, "312", False), True),
    ],
)
def test_any_undefined(values, result):
    assert undefined.any_undefined(*values) is result


@pytest.mark.parametrize(
    ("values", "result"),
    [
        ((undefined.UNDEFINED, undefined.UNDEFINED, "43123", "5421234"), 2),
        ((undefined.UNDEFINED, 233, 123, "4532123", 32123), 1),
        ((), 0),
        (("123", 543, "123", 321, "543", 123), 0),
        ((undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED, undefined.UNDEFINED), 5),
        ((undefined.UNDEFINED, "32123123", undefined.UNDEFINED, 34123123, undefined.UNDEFINED), 3),
    ],
)
def test_count(values, result):
    assert undefined.count(*values) == result



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Additional utilities for Nox."""

from __future__ import annotations

import os
import shutil

from pipelines import nox

DIRECTORIES_TO_DELETE = [
    ".nox",
    "build",
    "dist",
    "hikari.egg-info",
    "public",
    ".pytest_cache",
    ".mypy_cache",
    "node_modules",
]

FILES_TO_DELETE = [".coverage", "package-lock.json"]

TO_DELETE = [(shutil.rmtree, DIRECTORIES_TO_DELETE), (os.remove, FILES_TO_DELETE)]


@nox.session(venv_backend="none")
def purge(session: nox.Session) -> None:
    """Delete any nox-generated files."""
    for func, trash_list in TO_DELETE:
        for trash in trash_list:
            try:
                func(trash)
            except Exception as exc:  # noqa: BLE001, PERF203
                session.warn(f"[ FAIL ] Failed to remove {trash!r}: {exc!s}")
            else:
                session.log(f"[  OK  ] Removed {trash!r}")



# Hikari Examples - A collection of examples for Hikari.
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""A simple bot with only a `!ping` command."""

from __future__ import annotations

import os

import hikari

bot = hikari.GatewayBot(token=os.environ["BOT_TOKEN"])


@bot.listen()
async def on_message(event: hikari.MessageCreateEvent) -> None:
    """Listen for messages being created."""
    if not event.is_human:
        # Do not respond to bots or webhooks!
        return

    if event.content == "!ping":
        await event.message.respond(f"Pong! {bot.heartbeat_latency * 1_000:.0f}ms")


bot.run()
